import os
import signal
import logging
from datetime import datetime

from landscape.lib.process import ProcessInformation
from landscape.manager.plugin import ManagerPlugin


class ProcessNotFoundError(Exception):
    pass


class ProcessMismatchError(Exception):
    pass


class SignalProcessError(Exception):
    pass


class ProcessKiller(ManagerPlugin):
    """
    A management plugin that signals processes upon receiving a message from
    the server.
    """

    def __init__(self, process_info=None):
        if process_info is None:
            process_info = ProcessInformation()
        self.process_info = process_info

    def register(self, registry):
        super(ProcessKiller, self).register(registry)
        registry.register_message("signal-process",
                                  self._handle_signal_process)

    def _handle_signal_process(self, message):
        self.call_with_operation_result(message, self.signal_process,
                                        message["pid"], message["name"],
                                        message["start-time"],
                                        message["signal"])

    def signal_process(self, pid, name, start_time, signame):
        logging.info("Sending %s signal to the process with PID %d.",
                     signame, pid)
        process_info = self.process_info.get_process_info(pid)
        if not process_info:
            start_time = datetime.utcfromtimestamp(start_time)
            message = ("The process %s with PID %d that started at %s UTC was "
                       "not found") % (name, pid, start_time)
            raise ProcessNotFoundError(message)
        elif abs(process_info["start-time"] - start_time) > 2:
            # We don't check that the start time matches precisely because
            # the way we obtain boot times isn't very precise, and this may
            # cascade into having imprecise process start times.
            expected_time = datetime.utcfromtimestamp(start_time)
            actual_time = datetime.utcfromtimestamp(process_info["start-time"])
            message = ("The process %s with PID %d that started at "
                       "%s UTC was not found.  A process with the same "
                       "PID that started at %s UTC was found and not "
                       "sent the %s signal") % (name, pid, expected_time,
                                                actual_time, signame)
            raise ProcessMismatchError(message)

        signum = getattr(signal, "SIG%s" % (signame,))
        try:
            os.kill(pid, signum)
        except:
            # XXX Nothing is indicating what the problem was.
            message = ("Attempting to send the %s signal to the process "
                       "%s with PID %d failed") % (signame, name, pid)
            raise SignalProcessError(message)



import glob
import os
import pwd
import grp
import shutil
import tempfile

from twisted.internet.defer import succeed

from landscape.lib.twisted_util import spawn_process

from landscape.manager.plugin import ManagerPlugin
from landscape.package.reporter import find_reporter_command


class ProcessError(Exception):
    """Exception raised when running a process fails."""


class AptSources(ManagerPlugin):
    """A plugin managing sources.list content."""

    SOURCES_LIST = "/etc/apt/sources.list"
    SOURCES_LIST_D = "/etc/apt/sources.list.d"

    def register(self, registry):
        super(AptSources, self).register(registry)
        registry.register_message("apt-sources-replace",
                                  self._handle_repositories)

    def _run_process(self, command, args, uid=None, gid=None):
        """
        Run the process in an asynchronous fashion, to be overriden in tests.
        """
        return spawn_process(command, args, uid=uid, gid=gid)

    def _handle_process_error(self, result):
        """
        Turn a failed process command (code != 0) to a C{ProcessError}.
        """
        out, err, code = result
        if code:
            raise ProcessError("%s\n%s" % (out, err))

    def _handle_process_failure(self, failure):
        """
        Turn a signaled process command to a C{ProcessError}.
        """
        if not failure.check(ProcessError):
            out, err, signal = failure.value
            raise ProcessError("%s\n%s" % (out, err))
        else:
            return failure

    def _remove_and_continue(self, passthrough, path):
        """
        Remove the temporary file created for the process, and forward the
        result.
        """
        os.unlink(path)
        return passthrough

    def _handle_repositories(self, message):
        """
        Handle a list of repositories to set on the machine.

        The format is the following:

        {"sources": [
          {"name": "repository-name",
           "content":
              "deb http://archive.ubuntu.com/ubuntu/ maverick main\n\
              "deb-src http://archive.ubuntu.com/ubuntu/ maverick main"}
          {"name": "repository-name-dev",
           "content":
              "deb http://archive.ubuntu.com/ubuntu/ maverick universe\n\
              "deb-src http://archive.ubuntu.com/ubuntu/ maverick universe"}],
         "gpg-keys": ["-----BEGIN PGP PUBLIC KEY BLOCK-----\n\
                      XXXX
                      -----END PGP PUBLIC KEY BLOCK-----",
                      "-----BEGIN PGP PUBLIC KEY BLOCK-----\n\
                      YYY
                      -----END PGP PUBLIC KEY BLOCK-----"]}
        """
        deferred = succeed(None)
        for key in message["gpg-keys"]:
            fd, path = tempfile.mkstemp()
            os.close(fd)
            key_file = file(path, "w")
            key_file.write(key)
            key_file.close()
            deferred.addCallback(
                lambda ignore, path=path:
                    self._run_process("/usr/bin/apt-key", ["add", path]))
            deferred.addCallback(self._handle_process_error)
            deferred.addBoth(self._remove_and_continue, path)
        deferred.addErrback(self._handle_process_failure)
        deferred.addCallback(self._handle_sources, message["sources"])
        return self.call_with_operation_result(message, lambda: deferred)

    def _handle_sources(self, ignored, sources):
        """Handle sources repositories."""
        fd, path = tempfile.mkstemp()
        os.close(fd)
        new_sources = file(path, "w")
        for line in file(self.SOURCES_LIST):
            stripped_line = line.strip()
            if not stripped_line or stripped_line.startswith("#"):
                new_sources.write(line)
            else:
                new_sources.write("#%s" % line)
        new_sources.close()

        original_stat = os.stat(self.SOURCES_LIST)
        shutil.move(path, self.SOURCES_LIST)
        os.chmod(self.SOURCES_LIST, original_stat.st_mode)
        os.chown(self.SOURCES_LIST, original_stat.st_uid, original_stat.st_gid)

        for filename in glob.glob(os.path.join(self.SOURCES_LIST_D, "*.list")):
            shutil.move(filename, "%s.save" % filename)

        for source in sources:
            filename = os.path.join(self.SOURCES_LIST_D,
                                    "landscape-%s.list" % source["name"])
            sources_file = file(filename, "w")
            sources_file.write(source["content"])
            sources_file.close()
            os.chmod(filename, 0644)
        return self._run_reporter().addCallback(lambda ignored: None)

    def _run_reporter(self):
        """Once the repositories are modified, trigger a reporter run."""
        reporter = find_reporter_command()

        # Force an apt-update run, because the sources.list has changed
        args = ["--force-apt-update"]

        if self.registry.config.config is not None:
            args.append("--config=%s" % self.registry.config.config)

        if os.getuid() == 0:
            uid = pwd.getpwnam("landscape").pw_uid
            gid = grp.getgrnam("landscape").gr_gid
        else:
            uid = None
            gid = None
        return self._run_process(reporter, args, uid=uid, gid=gid)



from string import Template
import pwd
import grp
import os


class BootstrapList(object):

    def __init__(self, bootstraps):
        self._bootstraps = bootstraps

    def bootstrap(self, **vars):
        for bootstrap in self._bootstraps:
            bootstrap.bootstrap(**vars)


class BootstrapPath(object):

    def __init__(self, path, username=None, group=None, mode=None):
        self.path = path
        self.username = username
        self.group = group
        self.mode = mode

    def _create(self, path):
        pass

    def bootstrap(self, **vars):
        path = Template(self.path).substitute(**vars)
        self._create(path)

        if self.mode is not None:
            os.chmod(path, self.mode)

        if os.getuid() == 0:
            if self.username is not None:
                uid = pwd.getpwnam(self.username).pw_uid
            else:
                uid = -1

            if self.group is not None:
                gid = grp.getgrnam(self.group).gr_gid
            else:
                gid = -1

            if uid != -1 or gid != -1:
                os.chown(path, uid, gid)


class BootstrapFile(BootstrapPath):

    def _create(self, path):
        open(path, "a").close()


class BootstrapDirectory(BootstrapPath):

    def _create(self, path):
        try:
            os.makedirs(path)
        except OSError:
            if not os.path.isdir(path):
                raise



import os

from landscape.tests.helpers import LandscapeTest

from landscape.lib.bootstrap import (
    BootstrapPath, BootstrapFile, BootstrapDirectory, BootstrapList)


class BootstrapPathTest(LandscapeTest):

    bootstrap_class = BootstrapPath

    def setUp(self):
        super(BootstrapPathTest, self).setUp()
        self.dirname = self.makeDir()
        self.path = os.path.join(self.dirname, "$my_var")
        self.real_path = os.path.join(self.dirname, "my_var_value")

    def test_username(self):
        getpwnam = self.mocker.replace("pwd.getpwnam")
        getpwnam("username").pw_uid
        self.mocker.result(1234)

        getuid = self.mocker.replace("os.getuid")
        getuid()
        self.mocker.result(0)

        chown = self.mocker.replace("os.chown")
        chown(self.real_path, 1234, -1)

        self.mocker.replay()

        file = self.bootstrap_class(self.real_path, username="username")
        file.bootstrap(my_var="my_var_value")

    def test_group(self):
        getgrnam = self.mocker.replace("grp.getgrnam")
        getgrnam("group").gr_gid
        self.mocker.result(5678)

        getuid = self.mocker.replace("os.getuid")
        getuid()
        self.mocker.result(0)

        chown = self.mocker.replace("os.chown")
        chown(self.real_path, -1, 5678)

        self.mocker.replay()

        file = self.bootstrap_class(self.path, group="group")
        file.bootstrap(my_var="my_var_value")

    def test_mode(self):
        chmod = self.mocker.replace("os.chmod")
        chmod(self.real_path, 0644)

        self.mocker.replay()

        file = self.bootstrap_class(self.path, mode=0644)
        file.bootstrap(my_var="my_var_value")

    def test_all_details(self):
        getuid = self.mocker.replace("os.getuid")
        getuid()
        self.mocker.result(0)

        getpwnam = self.mocker.replace("pwd.getpwnam")
        getpwnam("username").pw_uid
        self.mocker.result(1234)

        getgrnam = self.mocker.replace("grp.getgrnam")
        getgrnam("group").gr_gid
        self.mocker.result(5678)

        chown = self.mocker.replace("os.chown")
        chown(self.real_path, 1234, 5678)

        chmod = self.mocker.replace("os.chmod")
        chmod(self.real_path, 0644)

        self.mocker.replay()

        file = self.bootstrap_class(self.path, "username", "group", 0644)
        file.bootstrap(my_var="my_var_value")

    def test_all_details_with_non_root(self):
        getuid = self.mocker.replace("os.getuid")
        getuid()
        self.mocker.result(1000)

        chmod = self.mocker.replace("os.chmod")
        chmod(self.real_path, 0644)

        self.mocker.replay()

        file = self.bootstrap_class(self.path, "username", "group", 0644)
        file.bootstrap(my_var="my_var_value")


class BootstrapCreationTest(BootstrapPathTest):

    bootstrap_class = BootstrapFile

    def exists(self, path):
        return os.path.isfile(path)

    def test_creation(self):
        file = self.bootstrap_class(self.path)
        self.assertFalse(self.exists(self.real_path))
        file.bootstrap(my_var="my_var_value")
        self.assertTrue(self.exists(self.real_path))


class BootstrapFileTest(BootstrapCreationTest):

    def test_creation_wont_overwrite(self):
        filename = self.makeFile("CONTENT")
        file = self.bootstrap_class(filename)
        file.bootstrap()
        self.assertEqual(open(filename).read(), "CONTENT")


class BootstrapDirectoryTest(BootstrapCreationTest):

    bootstrap_class = BootstrapDirectory

    def exists(self, path):
        return os.path.isdir(path)

    def test_creation_works_with_existing(self):
        dirname = self.makeDir()
        dir = self.bootstrap_class(dirname)
        dir.bootstrap()
        self.assertTrue(self.exists(dirname))

    def test_creation_will_fail_correctly(self):
        filename = self.makeFile("I AM A *FILE*")
        dir = self.bootstrap_class(filename)
        self.assertRaises(OSError, dir.bootstrap)


class BootstrapListTest(LandscapeTest):

    def test_creation(self):
        dirname = self.makeDir()

        list = BootstrapList([BootstrapFile("$dirname/filename"),
                              BootstrapDirectory("$dirname/dirname"),
                              BootstrapFile("$dirname/dirname/filename")])

        list.bootstrap(dirname=dirname)

        self.assertTrue(os.path.isfile(os.path.join(dirname, "filename")))
        self.assertTrue(os.path.isdir(os.path.join(dirname, "dirname")))
        self.assertTrue(os.path.isfile(os.path.join(dirname,
                                                    "dirname/filename")))



from landscape.lib.cloud import (
    EC2_API, _fetch_ec2_item, fetch_ec2_meta_data, MAX_LENGTH)
from landscape.lib.fetch import HTTPCodeError, PyCurlError
from landscape.tests.helpers import LandscapeTest
from twisted.internet.defer import succeed, fail


class CloudTest(LandscapeTest):

    def setUp(self):
        LandscapeTest.setUp(self)
        self.query_results = {}
        self.kwargs = {}

        def fetch_stub(url, **kwargs):
            self.kwargs = kwargs
            value = self.query_results[url]
            if isinstance(value, Exception):
                return fail(value)
            else:
                return succeed(value)

        self.fetch_func = fetch_stub
        self.add_query_result("instance-id", "i00001")
        self.add_query_result("ami-id", "ami-00002")
        self.add_query_result("instance-type", "hs1.8xlarge")

    def add_query_result(self, name, value):
        """
        Add a url to self.query_results that is then available through
        self.fetch_func.
        """
        url = "%s/meta-data/%s" % (EC2_API, name)
        self.query_results[url] = value

    def test_fetch_ec2_meta_data_error_on_any_item_error(self):
        """
        L{_fetch_ec2_meta_data} returns a deferred C{Failure} containing the
        error message when an error occurs on any of the queried meta-data
        items C{instance-id}, C{ami-id} or C{instance-type}.
        """
        self.log_helper.ignore_errors(HTTPCodeError)
        error = HTTPCodeError(404, "notfound")
        metadata_items = ["instance-id", "ami-id", "instance-type"]
        for item in metadata_items:
            # reset all item data adding the error to only 1 item per iteration
            for setup_item in metadata_items:
                if setup_item == item:
                    self.add_query_result(item, error)
                else:
                    self.add_query_result(setup_item, "value%s" % setup_item)

            deferred = fetch_ec2_meta_data(fetch=self.fetch_func)
            failure = self.failureResultOf(deferred)
            self.assertEqual(
                "Server returned HTTP code 404",
                failure.getErrorMessage())

    def test_fetch_ec2_meta_data(self):
        """
        L{_fetch_ec2_meta_data} returns a C{dict} containing meta-data for
        C{instance-id}, C{ami-id} and C{instance-type}.
        """
        deferred = fetch_ec2_meta_data(fetch=self.fetch_func)
        result = self.successResultOf(deferred)
        self.assertEqual(
            {"ami-id": u"ami-00002",
             "instance-id": u"i00001",
             "instance-type": u"hs1.8xlarge"},
            result)

    def test_fetch_ec2_meta_data_utf8(self):
        """
        L{_fetch_ec2_meta_data} decodes utf-8 strings returned from the
        external service.
        """
        self.add_query_result("ami-id", "asdf\xe1\x88\xb4")
        deferred = fetch_ec2_meta_data(fetch=self.fetch_func)
        result = self.successResultOf(deferred)
        self.assertEqual({"instance-id": u"i00001",
                          "ami-id": u"asdf\u1234",
                          "instance-type": u"hs1.8xlarge"},
                         result)

    def test_fetch_ec2_meta_data_truncates(self):
        """L{_fetch_ec2_meta_data} truncates values that are too long."""
        self.add_query_result("ami-id", "a" * MAX_LENGTH * 5)
        self.add_query_result("instance-id", "b" * MAX_LENGTH * 5)
        self.add_query_result("instance-type", "c" * MAX_LENGTH * 5)
        deferred = fetch_ec2_meta_data(fetch=self.fetch_func)
        result = self.successResultOf(deferred)
        self.assertEqual(
            {"ami-id": "a" * MAX_LENGTH,
             "instance-id": "b" * MAX_LENGTH,
             "instance-type": "c" * MAX_LENGTH},
            result)

    def test_wb_fetch_ec2_item_multiple_items_appends_accumulate_list(self):
        """
        L{_fetch_ec2_item} retrieves individual meta-data items from the
        EC2 api and appends them to the C{list} provided by the C{accumulate}
        parameter.
        """
        accumulate = []
        self.successResultOf(
            _fetch_ec2_item("instance-id", accumulate, fetch=self.fetch_func))
        self.successResultOf(
            _fetch_ec2_item(
                "instance-type", accumulate, fetch=self.fetch_func))
        self.assertEqual(["i00001", "hs1.8xlarge"], accumulate)

    def test_wb_fetch_ec2_item_error_returns_failure(self):
        """
        L{_fetch_ec2_item} returns a deferred C{Failure} containing the error
        message when faced with no EC2 cloud API service.
        """
        self.log_helper.ignore_errors(PyCurlError)
        self.add_query_result("other-id", PyCurlError(60, "pycurl error"))
        accumulate = []
        deferred = _fetch_ec2_item(
            "other-id", accumulate, fetch=self.fetch_func)
        failure = self.failureResultOf(deferred)
        self.assertEqual("Error 60: pycurl error", failure.getErrorMessage())

    def test_wb_fetch_ec2_meta_data_nofollow(self):
        """
        L{_fetch_ec2_meta_data} sets C{follow} to C{False} to avoid following
        HTTP redirects.
        """
        self.log_helper.ignore_errors(PyCurlError)
        self.add_query_result("other-id", PyCurlError(60, "pycurl error"))
        accumulate = []
        deferred = _fetch_ec2_item(
            "other-id", accumulate, fetch=self.fetch_func)
        self.failureResultOf(deferred)
        self.assertEqual({"follow": False}, self.kwargs)



"""
The accumulation logic generates data points for times that are a
multiple of a step size.  In other words, if the step size is 300
seconds, any data reported by the accumulation code will always be for
a timestamp that is a multiple of 300.  The purpose of this behaviour
is to (a) limit the amount of data that is sent to the server and (b)
provide data in a predictable format to make server-side handling of
the data straight-forward.  A nice side-effect of providing data at a
known step-interval is that the server can detect blackholes in the
data simply by testing for the absence of data points at step
intervals.

Limiting the amount of data sent to the server and making the data
format predictable are both desirable attributes, but we need to
ensure the data reported is accurate.  We can't rely on plugins to
report data exactly at step boundaries and even if we could we
wouldn't necessarily end up with data points that are representative
of the resource being monitored.  We need a way to calculate a
representative data point from the set of data points that a plugin
provided during a step period.

Suppose we want to calculate data points for timestamps 300 and 600.
Assume a plugin runs at an interval less than 300 seconds to get
values to provide to the accumulator.  Each value received by the
accumulator is used to update a data point that will be sent to the
server when we cross the step boundary.  The algorithm, based on
derivatives, is:

  (current time - previous time) * value + last accumulated value

If the 'last accumulated value' isn't available, it defaults to 0.
For example, consider these timestamp/load average measurements:
300/2.0, 375/3.0, 550/3.5 and 650/0.5.  Also assume we have no data
prior to 300/2.0.  This data would be processed as follows:

  Input       Calculation                   Accumulated Value
  -----       -----------                   -----------------
  300/2.0     (300 - 300) * 2.0 + 0         0.0

  375/3.0     (375 - 300) * 3.0 + 0.0       225.0

  550/3.5     (550 - 375) * 3.5 + 225.0     837.5

  650/0.5     (600 - 550) * 0.5 + 837.5     862.5

Notice that the last value crosses a step boundary; the calculation
for this value is:

  (step boundary time - previous time) * value + last accumulated value

This yields the final accumulated value for the step period we've just
traversed.  The data point sent to the server is generated using the
following calculation:

  accumulated value / step interval size

The data point sent to the server in our example would be:

  862.5 / 300 = 2.875

This value is representative of the activity that actually occurred
and is returned to the plugin to queue for delivery to the server.
The accumulated value for the next interval is calculated using the
portion of time that crossed into the new step period:

  Input       Calculation                   Accumulated Value
  -----       -----------                   -----------------
  650/0.5     (650 - 600) * 0.5 + 0         25

And so the logic goes, continuing in a similar fashion, yielding
representative data at each step boundary.
"""

class Accumulator(object):

    def __init__(self, persist, step_size):
        self._persist = persist
        self._step_size = step_size

    def __call__(self, new_timestamp, new_free_space, key):
        previous_timestamp, accumulated_value = self._persist.get(key, (0, 0))
        accumulated_value, step_data = \
            accumulate(previous_timestamp, accumulated_value,
                       new_timestamp, new_free_space, self._step_size)
        self._persist.set(key, (new_timestamp, accumulated_value))
        return step_data


def accumulate(previous_timestamp, accumulated_value,
               new_timestamp, new_value,
               step_size):
    previous_step = previous_timestamp // step_size
    new_step = new_timestamp // step_size
    step_boundary = new_step * step_size
    step_diff = new_step - previous_step
    step_data = None

    if step_diff == 0:
        diff = new_timestamp - previous_timestamp
        accumulated_value += diff * new_value
    elif step_diff == 1:
        diff = step_boundary - previous_timestamp
        accumulated_value += diff * new_value
        step_value = float(accumulated_value) / step_size
        step_data = (step_boundary, step_value)
        diff = new_timestamp - step_boundary
        accumulated_value = diff * new_value
    else:
        diff = new_timestamp - step_boundary
        accumulated_value = diff * new_value

    return accumulated_value, step_data



import base64
import os
import textwrap
import time

import apt_inst
import apt_pkg

from landscape.lib.fs import append_file, create_file
from landscape.package.facade import AptFacade


class AptFacadeHelper(object):
    """Helper that sets up an AptFacade with a tempdir as its root."""

    def set_up(self, test_case):
        test_case.apt_root = test_case.makeDir()
        self.dpkg_status = os.path.join(
            test_case.apt_root, "var", "lib", "dpkg", "status")
        test_case.Facade = AptFacade
        test_case.facade = AptFacade(root=test_case.apt_root)
        test_case.facade.refetch_package_index = True
        test_case._add_system_package = self._add_system_package
        test_case._install_deb_file = self._install_deb_file
        test_case._add_package_to_deb_dir = self._add_package_to_deb_dir
        test_case._touch_packages_file = self._touch_packages_file
        test_case._hash_packages_by_name = self._hash_packages_by_name

    def _add_package(self, packages_file, name, architecture="all",
                     version="1.0", control_fields=None):
        if control_fields is None:
            control_fields = {}
        package_stanza = textwrap.dedent("""
                Package: %(name)s
                Priority: optional
                Section: misc
                Installed-Size: 1234
                Maintainer: Someone
                Architecture: %(architecture)s
                Source: source
                Version: %(version)s
                Description: description
                """ % {"name": name, "version": version,
                       "architecture": architecture})
        package_stanza = apt_pkg.rewrite_section(
            apt_pkg.TagSection(package_stanza), apt_pkg.REWRITE_PACKAGE_ORDER,
            control_fields.items())
        append_file(packages_file, "\n" + package_stanza + "\n")

    def _add_system_package(self, name, architecture="all", version="1.0",
                            control_fields=None):
        """Add a package to the dpkg status file."""
        system_control_fields = {"Status": "install ok installed"}
        if control_fields is not None:
            system_control_fields.update(control_fields)
        self._add_package(
            self.dpkg_status, name, architecture=architecture, version=version,
            control_fields=system_control_fields)

    def _install_deb_file(self, path):
        """Fake the the given deb file is installed in the system."""
        deb_file = open(path)
        deb = apt_inst.DebFile(deb_file)
        control = deb.control.extractdata("control")
        deb_file.close()
        lines = control.splitlines()
        lines.insert(1, "Status: install ok installed")
        status = "\n".join(lines)
        append_file(self.dpkg_status, status + "\n\n")

    def _add_package_to_deb_dir(self, path, name, architecture="all",
                                version="1.0", control_fields=None):
        """Add fake package information to a directory.

        There will only be basic information about the package
        available, so that get_packages() have something to return.
        There won't be an actual package in the dir.
        """
        if control_fields is None:
            control_fields = {}
        self._add_package(
            os.path.join(path, "Packages"), name, architecture=architecture,
            version=version, control_fields=control_fields)

    def _touch_packages_file(self, deb_dir):
        """Make sure the Packages file get a newer mtime value.

        If we rely on simply writing to the file to update the mtime, we
        might end up with the same as before, since the resolution is
        seconds, which causes apt to not reload the file.
        """
        packages_path = os.path.join(deb_dir, "Packages")
        mtime = int(time.time() + 1)
        os.utime(packages_path, (mtime, mtime))

    def _hash_packages_by_name(self, facade, store, package_name):
        """
        Ensure the named L{Package} is correctly recorded in the store so that
        we can really test the functions of the facade that depend on it.
        """
        hash_ids = {}
        for version in facade.get_packages_by_name(package_name):
            skeleton = facade.get_package_skeleton(
                version, with_info=False)
            hash = skeleton.get_hash()
            facade._pkg2hash[(version.package, version)] = hash
            hash_ids[hash] = version.package.id
        store.set_hash_ids(hash_ids)


class SimpleRepositoryHelper(object):
    """Helper for adding a simple repository to the facade.

    This helper requires that C{test_case.facade} has already been set
    up.
    """

    def set_up(self, test_case):
        test_case.repository_dir = test_case.makeDir()
        create_simple_repository(test_case.repository_dir)
        test_case.facade.add_channel_deb_dir(test_case.repository_dir)


PKGNAME1 = "name1_version1-release1_all.deb"
PKGNAME2 = "name2_version2-release2_all.deb"
PKGNAME3 = "name3_version3-release3_all.deb"
PKGNAME4 = "name3_version3-release4_all.deb"
PKGNAME_MINIMAL = "minimal_1.0_all.deb"
PKGNAME_SIMPLE_RELATIONS = "simple-relations_1.0_all.deb"
PKGNAME_VERSION_RELATIONS = "version-relations_1.0_all.deb"
PKGNAME_MULTIPLE_RELATIONS = "multiple-relations_1.0_all.deb"
PKGNAME_OR_RELATIONS = "or-relations_1.0_all.deb"

PKGDEB1 = ("ITxhcmNoPgpkZWJpYW4tYmluYXJ5ICAgMTE2NjExNDQ5MyAgMCAgICAgMCAgICAgMT"
           "AwNjQ0ICA0ICAgICAgICAgYAoyLjAKY29udHJvbC50YXIuZ3ogIDExNjYxMTQ0OTMg"
           "IDAgICAgIDAgICAgIDEwMDY0NCAgNDUyICAgICAgIGAKH4sIAAAAAAAAA+3UQW+bMB"
           "QHcM58Ch+7QwCbEJpomzat0rTDpmiRenfNC7EGNnuGSOmnnwMlyVK1O6VT1feTkJ+e"
           "/wRh40RxcHGJl2dZP3rnY1/zJJsmIs9Fvs/5UQQsC15A51qJjAVobftc7l/zr1QU10"
           "XmutpdeP9n0+mT+8+5+Hv/fSOdBiyh/b84MYM1n2fz7G4t0+u5SvMkhbTgs3wu+CwB"
           "xjqHsdtIhLiwKjayBh6rjTQlVLaMbuBOSxOV92FAXuX5V9a0aKv/eP5zkZyf/1TQ+X"
           "8RS6l+yRIWrD/Y4S2g09Ys2HYo+AShAun81ApU2099Rds1PFyitqjb3YLZZj8hq/Az"
           "qo1ufa5D/4uyqnwIJjfQgCncgjUICL87jdA/jF19OGmND3wXHvLn4UfJn6BsXY/hsT"
           "7Jj63jLauuLMG1/gb3UB3iY+MY/mLNutJqn1ZjeYgfOsf8Eu1WF9C/6lANq/rN+I+s"
           "qqCYrPS9XxlxHX6X2rT+AvQLuv8Gt5b90FDDDpC9L4fOJ/PQiQy0H/3COIW6GXZh1d"
           "W1xB0P2Umb078wIYQQQgghhBBCCCGEEEIIIYS8UX8AYydx2gAoAABkYXRhLnRhci5n"
           "eiAgICAgMTE2NjExNDQ5MyAgMCAgICAgMCAgICAgMTAwNjQ0ICAzOTQgICAgICAgYA"
           "ofiwgAAAAAAAAD09NnoDkwAAJzU1MwDQToNJhtaGBqYmBkbm5kDlIHpI0YFEwZ6ABK"
           "i0sSixQUGIry80vwqSMkP0SBnn5pcZH+YIp/EwYDIMd4NP7pGP/FGYlFqfqDJ/4NzY"
           "xNRuOf3vGfkp+sPzji38jEwHA0/gci/vMSc1MN9Qc6/o2B7NH4H7j4T85IzEtPzclP"
           "13NJTcpMzNNLr6Iw/s1MTHDGv5GxOSz+zUxNjYDxbw7kMSgYjMY/zYF8NwdHVm2jKx"
           "Mzepwz6J7y5jpkIOH6sDKssF1rmUqYzBX2piZj9zyFad5RHv8dLoXsqua2spF3v+PQ"
           "ffXIlN8aYepsu3x2u0202VX+QFC10st6vvMfDdacgtdzKtpe5G5tuFYx5elcpXm27O"
           "d8LH7Oj3mqP7VgD8P6dTmJ33dsPnpuBnPO3SvLDNlu6ay9It6yZon0BIZRMApGwSgY"
           "BaNgFIyCUTAKRsEoGAWjYBSMglEwCkbBKBgFo2AUjIJRMApGAUkAADhX8vgAKAAA ")

PKGDEB2 = ("ITxhcmNoPgpkZWJpYW4tYmluYXJ5ICAgMTE2NjExNDUyMiAgMCAgICAgMCAgICAgMT"
           "AwNjQ0ICA0ICAgICAgICAgYAoyLjAKY29udHJvbC50YXIuZ3ogIDExNjYxMTQ1MjIg"
           "IDAgICAgIDAgICAgIDEwMDY0NCAgNDUyICAgICAgIGAKH4sIAAAAAAAAA+3UTY/TMB"
           "AG4JzzK3yEQ/Phxk1aAQKxEuIAqrYSd+NMU4vEDuOkUvfX4yabthQBpy5aMY9UZTR+"
           "06ieuFEc3Fzi5UIMV+/6OtRpIrKE5/l8zn0/z9MsYCJ4Ar3rJDIWoLXdn3J/W3+mor"
           "gphesbd+P5L7Lst/NPU/7z/H2DLwKW0PxvrixSlYkiAVGIxZJnaSHFdilUDplabnnG"
           "WO8wdjuJEJdWxUY2wGO1k6aC2lbRHXzV0kTVQxiQZ3n+lTUd2vofnv+cJ9fnf57S+X"
           "8Sa6m+yQpWbDjY4RdAp61Zsf1Y8BlCDdL5pQ2oblj6gLZvebhGbVF3hxWz7XFB1uE7"
           "VDvd+VyP/htlXfsQzO6gBVO6FWsREL73GmF4GHvx+qI1PfBleMpfh39J3oOyTTOFp/"
           "oiP7XOt2z6qgLX+RvcY3WKT41z+L0121qrY1pN5Sl+6pzza7R7XcLwU8dq3NWPxr9k"
           "dQ3lbKMf/M7wIvwkten8B9Bv6PEd3Fv2WUMDB0D2qho7b81jJzLQvfEb4xTqdpzCpm"
           "8aiQcesos2p39hQgghhBBCCCGEEEIIIYQQQgj5T/0AyM2cyQAoAABkYXRhLnRhci5n"
           "eiAgICAgMTE2NjExNDUyMiAgMCAgICAgMCAgICAgMTAwNjQ0ICAzOTMgICAgICAgYA"
           "ofiwgAAAAAAAAD09NnoDkwAAJzU1MwDQToNJhtaGBqYmBkbm5sbAgUNzc3NGZQMGWg"
           "AygtLkksUlBgKMrPL8GnjpD8EAV6+qXFRfqDLP6BHCOT0finX/wXZyQWpeoPnvg3ND"
           "MyG41/esd/Sn6y/uCIfyNj89Hyf0DiPy8xN9VIf6Dj39jY3HQ0/gcu/pMzEvPSU3Py"
           "0/VcUpMyE/P00qsojH8zExOc8Q/M7Yj4Bxb8BobmBsDkomAwGv80B/LdHBzX6hpdmZ"
           "jR45xB99RGrkMGEq4Pbf0L3UWDL4XIRIk6Hjx7Urzj6SSxS/YTzKbu28sqe/64oPmF"
           "JGPj3lqR1cLMdz12u04rLHp/gM2y0mv3HOc/GqxvCl7PqWh7kbux6VrFk69zlefZsu"
           "v5WPycH/NUv7VgF8N6vfeBcgXp3NlnBFNDw5eZsd1as/aK+JzyvZ0TGEbBKBgFo2AU"
           "jIJRMApGwSgYBaNgFIyCUTAKRsEoGAWjYBSMglEwCkbBKBgFJAEAu4OlKQAoAAAK")

PKGDEB3 = ("ITxhcmNoPgpkZWJpYW4tYmluYXJ5ICAgMTE2OTE0ODIwMyAgMCAgICAgMCAgICAgMT"
           "AwNjQ0ICA0ICAgICAgICAgYAoyLjAKY29udHJvbC50YXIuZ3ogIDExNjkxNDgyMDMg"
           "IDAgICAgIDAgICAgIDEwMDY0NCAgNDUxICAgICAgIGAKH4sIAAAAAAAAA+3UwY7TMB"
           "AG4JzzFD7CoUkax7iqYAViJcQBVFGJu3GmqbWJHcZJpe7T4yabtnS1cOqiFfNJVUbj"
           "P43qiZuk0dVlgRRiuAaX16GeZ0JwWRSF4KEvZc4jJqJn0PtOIWMROtf9Kfe39RcqSZ"
           "tS+L7xV57/m6J4cv7zef77/EODi4hlNP+r4yIrc1mUUs43C1VmhcxLEAKkFouCbzRj"
           "vcfUbxVCWjqdWtUAT/VW2QpqVyW38MMom1T3cURe5PnXznbo6n94/mWeXZ5/ntP5fx"
           "Yrpe9UBUs2HOz4O6A3zi7Zbiz4DKEG5cPSGnQ3LH1C17c8XqFxaLr9krn2sKDq+APq"
           "relCrsfwjaquQwhmt9CCLf2StQgIP3uDMDyMvXp31poe+Do+5i/Dj5LfQLummcJTfZ"
           "afWqdb1n1Vge/CDf6hOsanxin80dlNbfQhrafyGD92TvkVup0pYfipYzXu6mcbXrK6"
           "hnK2NvdhZ/JF/EUZ24UPYNjQwzu4c+yrgQb2gOxtNXbe24dOYqG7CRvjNZp2nMK6bx"
           "qFex6zszanf2FCCCGEEEIIIYQQQgghhBBCCPlP/QK+dA1dACgAAApkYXRhLnRhci5n"
           "eiAgICAgMTE2OTE0ODIwMyAgMCAgICAgMCAgICAgMTAwNjQ0ICAzOTkgICAgICAgYA"
           "ofiwgAAAAAAAAD09NnoDkwAAJzU1MwDQToNJhtaGBqamxuYmJiagQUNzc3MmJQMGWg"
           "AygtLkksUlBgKMrPL8GnjpD8EAV6+qXFRfqDKf4NGQyAHOPR+Kdj/BdnJBal6g+e+D"
           "c0MzYZjX96x39KfrL+4Ih/IxMDw9H4H4j4z0vMTTXWH8j4B9b/hsYmBqaj8T9w8Z+c"
           "kZiXnpqTn67nkpqUmZinl15FYfybmZjgjH8jY3NE/JuYAePfHKieQcFgNP5pDuS7OT"
           "jUTq53ZWJGj3MG3VPeXIcMJFwfVoYVtmstW+Imc4W9qcnYPU9hmneUx3+HSyG7qrmt"
           "bOTd7zh0Xz0y5bdGmDrbLp/dbhNtdpU/EFSt9LKe7/xHgzWn4PWcirYXuVsbrlVMeT"
           "pXaZ4t+zkfi5/zY57qTy3Yw7B+XU7g+8L07rmG7Fe2bVxmyHZLZ+0V8Sl2Xj8mMIyC"
           "UTAKRsEoGAWjYBSMglEwCkbBKBgFo2AUjIJRMApGwSgYBaNgFIyCUTAKSAIAY/FOKA"
           "AoAAAK")

PKGDEB4 = ("ITxhcmNoPgpkZWJpYW4tYmluYXJ5ICAgMTI3NjUxMTU3OC41MCAgICAgMCAgICAgNj"
           "Q0ICAgICA0\nICAgICAgICAgYAoyLjAKY29udHJvbC50YXIuZ3ogIDEyNzY1MTE1Nz"
           "guNTAgICAgIDAgICAgIDY0\nNCAgICAgMjk1ICAgICAgIGAKH4sIAFoFFkwC/+3TwU"
           "6EMBAGYM48RV9goS0dqnszMSbeTEy8F6iE\nCJS04MGnt2GzBzHqiVWT/7u0yVCm8G"
           "eyPNkdjzTRukbbdd0LoTgpLqmQCRdCckoYJRewhNn4eBXv\n3Pzdcz/Vtx/3T2R57c"
           "bZu37n/EulvsxfqnKTvyyFTBhH/rt7MPWLae2RjWawIn2yPnRuPLLX00Zk\n4uBtb0"
           "2Ixfsx/qu+t83hsXuLRwRPb22ofTfN65kbFsww9ZYtU+tNY9l0ennK7pxnsw1zN7bn"
           "YsjS\nD72LT72Lc2eVJrDb/A8NhWUIvzj/nMR2/kkKzP8lNERFJZWOGWiqiF89ayVt"
           "qbWhSlfimrEsD26w\nGEEAAAAAAAAAAAAAAAAAAIC/6x1piYqhACgAAApkYXRhLnRh"
           "ci5neiAgICAgMTI3NjUxMTU3OC41\nMCAgICAgMCAgICAgNjQ0ICAgICAxNDUgICAg"
           "ICAgYAofiwgAWgUWTAL/7dFBCsMgEEDRWfcUniCZ\nsU57kJ5ASJdFSOz9K9kULLQr"
           "C4H/NiPqQvnTLMNpc3XfZ9PPfW2W1JOae9s3i5okuPzBc6t5bU9Z\nS6nf7v067z93"
           "ENO8lcd9fP/LZ/d3f4td/6h+lqD0H+7W6ocl13wSAAAAAAAAAAAAAAAAAAfzAqr5\n"
           "GFYAKAAACg==\n")

PKGDEB_MINIMAL = (
    "ITxhcmNoPgpkZWJpYW4tYmluYXJ5ICAgMTMxNzg5MDQ3OSAgMCAgICAgMCAgICAgMTAwNj"
    "Q0ICA0 ICAgICAgICAgYAoyLjAKY29udHJvbC50YXIuZ3ogIDEzMTc4OTA0NzkgIDAgICA"
    "gIDAgICAgIDEw MDY0NCAgMjU4ICAgICAgIGAKH4sIAAAAAAACA+3Rz0rEMBAG8Jz7FPME"
    "3aT/FoqIC54EwZP3mB1s 1jQp0yz6+HaVBRFcTxWE7wfJQDKZHL5yo1anF9u2PVWzbfXXe"
    "qaM6Zq66pqurZQ2uqorRa36A8c5 WyFST4ck8ULfb/f/VLlxKWZJYeX8u6b5Mf+qbr7lb7"
    "rliDTyX92DdS/2mXsaffSjDcUjy+xT7MmU utiJG3xml4+ytNgQinvrY14WS093aYh0dVj"
    "2G36z4xS4dGm8Lm55duKn/DFmd55M0+dX9OrzQDHR nieOe47O80xJKOWBhYSDPb2cy0IB"
    "AAAAAAAAAAAAAAAAAAAAAMBF70s1/foAKAAAZGF0YS50YXIu Z3ogICAgIDEzMTc4OTA0N"
    "zkgIDAgICAgIDAgICAgIDEwMDY0NCAgMTA3ICAgICAgIGAKH4sIAAAA AAACA+3KsQ3CQB"
    "AEwCvlK4D/N4frMSGBkQz0jwmQiHCEo5lkpd09HOPv6mrMfGcbs37nR7R2Pg01"
    "ew5r32rvNUrGDp73x7SUEpfrbZl//LZ2AAAAAAAAAAAA2NELx33R7wAoAAAK")

PKGDEB_SIMPLE_RELATIONS = (
    "ITxhcmNoPgpkZWJpYW4tYmluYXJ5ICAgMTMxODUxNjMyMiAgMCAgICAgMCAgICAgMTAwNj"
    "Q0ICA0 ICAgICAgICAgYAoyLjAKY29udHJvbC50YXIuZ3ogIDEzMTg1MTYzMjIgIDAgICA"
    "gIDAgICAgIDEw MDY0NCAgMzQ0ICAgICAgIGAKH4sIAAAAAAACA+3R3UrDMBQH8F7nKc4L"
    "rGu2tYMi4tQrQRkI3mdp tNnSpKTZ0Lc37TYVQb2aIvx/0Ob09DQfPek4Obksmud5P/J5n"
    "n0cjxLOi1mez6ecT5KMZ5NJkVCe /IJtF4QnSlZr5+03dT+9/6fSsXQ2eGdO3P9iNvuy/3"
    "mWf+o/L6Y8oQz9P7mlkBvxpErqdNMaNfLK iKCd7diD8l0MSuJpxu6VDMNDozvJll47r8N"
    "LSa7t08KwhZe1DrFq6+NkwphYpEbXqlW26kpqvaqG mLO33DFx5eyj0TLElDyEnF16JTYx"
    "s+pHHidzO12pYaYh4uxWaBvipXxJN662dLaO9wv1LPqDpNI1 53GtTnrd7re+iJu3uhGG2"
    "v2hKdQiUC26w+Hp/fAU3Tna7f8BCa+OC1ekbfzwQ3HKEgAAAAAAAAAA AAAAAAAAAACAv/"
    "EKgcHt1gAoAABkYXRhLnRhci5neiAgICAgMTMxODUxNjMyMiAgMCAgICAgMCAg ICAgMTA"
    "wNjQ0ICAxMDcgICAgICAgYAofiwgAAAAAAAID7cqxDcJQEETBK8UVwH2b+64HQgIjGegf "
    "CJCIIMLRTPKC3d0+/i6f5qpX21z52bdorR+m7Fl9imw5jhVDxQbu19txHYY4nS/r8uX3aw"
    "cAAAAA AAAAAIANPQALnD6FACgAAAo=")


PKGDEB_VERSION_RELATIONS = (
    "ITxhcmNoPgpkZWJpYW4tYmluYXJ5ICAgMTMxODUxNjQ5OCAgMCAgICAgMCAgICAgMTAwNj"
    "Q0ICA0 ICAgICAgICAgYAoyLjAKY29udHJvbC50YXIuZ3ogIDEzMTg1MTY0OTggIDAgICA"
    "gIDAgICAgIDEw MDY0NCAgMzUwICAgICAgIGAKH4sIAAAAAAACA+3RQUvDMBQH8Jz7KXLU"
    "w7pmazcoczj1JAgDwXuW xTVbmpQkG/rtTds5RFBPGwj/H7R5fS9N07x0SM4ui6ZF0Y5sW"
    "mRfx0+EsUleFNNxznKSsWw0HhNa kAvY+8AdpWS1tc78Mu+v+j+VDoU1wVl95v5P8vzH/h"
    "eMfes/m7T9z9D/s1tyseMbWdKDdF5ZM3BS 8xADn7z0mZKyNEuepQjdQ628SJZOWafCe0l"
    "t06a5ThZOVCrEWXsXV+Nax0ly8CAbada+pI2T6y5m 9Gp2Q0dpdp2ciqfKsXBvzatWIsSS"
    "OIbta7O+euck38XSqh1jfj7v80tnD2otu491EUueuDIhXtKV 9NFWhs628X4r33jdaJkKW"
    "8/jLrxwqun/bhH/z6iaa9r0B0NDxQOtuKeng2n31C6qzObz1HyaEAAA AAAAAAAAAAAAAA"
    "AAAACAy/sAwTtOtwAoAABkYXRhLnRhci5neiAgICAgMTMxODUxNjQ5OCAgMCAg ICAgMCA"
    "gICAgMTAwNjQ0ICAxMDcgICAgICAgYAofiwgAAAAAAAID7cqxEcIwEETRK0UVgCT7UD0Q "
    "EpgZA/0DATNEEOHoveQHu7t9/F19GpmvtpH1s2/R2mGeemYfc9RW+9SjZGzgfr0d11LidL"
    "6sy5ff rx0AAAAAAAAAAAA29AD/ixlwACgAAAo=")


PKGDEB_MULTIPLE_RELATIONS = (
    "ITxhcmNoPgpkZWJpYW4tYmluYXJ5ICAgMTMxODU4MDA3OSAgMCAgICAgMCAgICAgMTAwNj"
    "Q0ICA0 ICAgICAgICAgYAoyLjAKY29udHJvbC50YXIuZ3ogIDEzMTg1ODAwNzkgIDAgICA"
    "gIDAgICAgIDEw MDY0NCAgMzgzICAgICAgIGAKH4sIAAAAAAACA+3RXUvDMBQG4F7nV5xL"
    "BVeb2nZQ5vDrShAGgvcx izaaNiXNRMEfb9atcwq6qwnC+8CW05N3bXcSH0d7lwTjPF+uf"
    "Jwn2+sg4rzI8nERYjxKeJJmaUR5 9AcWnReOKLp/sq75Jbdr/5+Kj6VtvLNmz+dfZNmP51"
    "+cZN/OnxdhmxKc/97NhHwWj6qkemG8bo0a OWWE17bp2J1yXShK4nHCbpX0/UWtO8lmTlu"
    "n/VtJtl22hWHnTlbah9TChdsJY0JIja5Uq5p5V1Lr 1LyvOR1MTimNk8Ojz2bKNsFNagit"
    "Gif0vq4yOphOv+yl7NI2D0ZLH34v1+XyOZN1bOil7MIp8RxS 98uVb92pb6Thne2Lnqv+h"
    "fuKHw1Vym6Ebnz4KFfSta0amjyF7zP1KuowuVjaehr+RyedblezOg/T anQtDLWrOZOvhK"
    "dKdJt504swC9XRg3WkhKxomH/MIgAAAAAAAAAAAAAAAAAAAACAHT4AFDs6bAAo AAAKZGF"
    "0YS50YXIuZ3ogICAgIDEzMTg1ODAwNzkgIDAgICAgIDAgICAgIDEwMDY0NCAgMTA3ICAg "
    "ICAgIGAKH4sIAAAAAAACA+3KsRHCMBBE0StFFYBkfFY9EBKYGWP3DwTMEEGEo/eSH+wejv"
    "F39aln vtp61s++RWvTeBpy6tmjtjqMLUrGDrb7el5Kicv1tsxffr92AAAAAAAAAAAA2NE"
    "Db6L1AQAoAAAK")


PKGDEB_OR_RELATIONS = (
    "ITxhcmNoPgpkZWJpYW4tYmluYXJ5ICAgMTMxNzg4ODg2OSAgMCAgICAgMCAgICAgMTAwNj"
    "Q0ICA0 ICAgICAgICAgYAoyLjAKY29udHJvbC50YXIuZ3ogIDEzMTc4ODg4NjkgIDAgICA"
    "gIDAgICAgIDEw MDY0NCAgMzc1ICAgICAgIGAKH4sIAAAAAAACA+3R30vDMBAH8D73r7g3"
    "FbS23brBUFHwSRAFwfeY HmtmlpQ08wf4x3vbWB2Cig8ThO8H2qbp3fXCZcfJzuViXFXLZ"
    "zGu8u3nRlIUo+GgHBXVUPaLvCwH CVXJH1h0UQWi5GHmg/sm7qfv/1R2rL2Lwdsdz380HH"
    "45f5n6p/kXo0GeUI7579yt0o9qyhPy4Siw VdF416X3HDpZTKjI8vSOdVy9zE2n09tgfDD"
    "xVTLa5bay6UXQjYkStQhSSFkrQXx0yS27uptQG7he rQvaPzmlMssP6O1jt0z7yD6sj9qE"
    "XCvjolwcJnTlG0cnM7mf84uat5Yz7ednUqbTwbTrXi+kW2fm ylK7PiHFRkVqVCcnpf6kW"
    "UrixtlX2uqZlKupXwcm47Rd1FwfUidLJh8b3qqyqr2qpJWTfzyxtC55 bi/2qfTcsJPvVi"
    "+WWW4qSdw3J301WZoAAAAAAAAAAAAAAAAAAAAAAPzOO2wqjioAKAAACmRhdGEu dGFyLmd"
    "6ICAgICAxMzE3ODg4ODY5ICAwICAgICAwICAgICAxMDA2NDQgIDEwNyAgICAgICBgCh+L "
    "CAAAAAAAAgPtyrsRwjAURNFXiioAfZBcjwkJzIyB/oGAGSIc4eic5Aa7h2P8XX6Zen+3TD"
    "1/9yNK"
    "GadWR2ltRC651hGpxw4et/u8phTny3Vdfvy2dgAAAAAAAAAAANjRE6Lr2rEAKAAACg==")


HASH1 = base64.decodestring("/ezv4AefpJJ8DuYFSq4RiEHJYP4=")
HASH2 = base64.decodestring("glP4DwWOfMULm0AkRXYsH/exehc=")
HASH3 = base64.decodestring("NJM05mj86veaSInYxxqL1wahods=")
HASH4 = 'c\xc1\xe6\xe1U\xde\xb6:\x03\xcb\xb9\xdc\xee\x91\xb7"\xc9\xb1\xe4\x8f'
HASH5 = '|\x93K\xe0gx\xba\xe4\x85\x84\xd9\xf4%\x8bB\xbdR\x97\xdb\xfc'
HASH6 = '\xedt!=,\\\rk\xa7\xe3$\xfb\x06\x9c\x88\x92)\xc2\xfb\xd6'
HASH7 = 'D\xb1\xb6\xf5\xaa\xa8i\x84\x07#x\x97\x01\xf7`.\x9b\xde\xfb '
HASH_MINIMAL = "6\xce\x8f\x1bM\x82MWZ\x1a\xffjAc(\xdb(\xa1\x0eG"
HASH_SIMPLE_RELATIONS = (
    "'#\xab&k\xe6\xf5E\xcfB\x9b\xceO7\xe6\xec\xa9\xddY\xaa")
HASH_VERSION_RELATIONS = (
    '\x84\xc9\xb4\xb3\r\x95\x16\x03\x95\x98\xc0\x14u\x06\xf7eA\xe65\xd1')
HASH_MULTIPLE_RELATIONS = (
    '\xec\xcdi\xdc\xde-\r\xc3\xd3\xc9s\x84\xe4\xc3\xd6\xc4\x12T\xa6\x0e')
HASH_OR_RELATIONS = (
    '\xa1q\xf4*\x1c\xd4L\xa1\xca\xf1\xfa?\xc3\xc7\x9f\x88\xd53B\xc9')

RELEASES = {"hardy": """Origin: Ubuntu
Label: Ubuntu
Codename: hardy
Version: 8.04
Date: Tue, 31 Mar 2009 13:30:02 +0000
Architectures: amd64 i386
Components: main restricted
MD5Sum:
 356312bc1c0ab2b8dbe5c67f09879497 827 main/binary-i386/Packages
 ad2d9b94381264ce25cda7cfa1b2da03 555 main/binary-i386/Packages.gz
 2f6ee66ed2d2b4115fabc8eed428e42e 78 main/binary-i386/Release
 f0fd5c1bb18584cf07f9bf4a9f2e6d92 605 main/binary-amd64/Packages
 98860034ca03a73a9face10af8238a81 407 main/binary-amd64/Packages.gz
 7e40db962fe49b6db232bf559cf6f79d 79 main/binary-amd64/Release
 99e2e7213a7fdd8dd6860623bbf700e6 538 restricted/binary-i386/Packages
 7771307958f2800bafb5cd96292308bd 384 restricted/binary-i386/Packages.gz
 8686ad9c5d83484dc66a1eca2bd8030f 84 restricted/binary-i386/Release
 99e2e7213a7fdd8dd6860623bbf700e6 538 restricted/binary-amd64/Packages
 7771307958f2800bafb5cd96292308bd 384 restricted/binary-amd64/Packages.gz
 6e24798a6089cd3a21226182784995e9 85 restricted/binary-amd64/Release
SHA1:
 1f39494284f8da4a1cdd788a3d91a048c5edf7f5 827 main/binary-i386/Packages
 e79a66d7543f24f77a9ffe1409431ae717781375 555 main/binary-i386/Packages.gz
 5fe86036c60d6210b662df3acc238e2936f03581 78 main/binary-i386/Release
 37ba69be70f4a79506038c0124293187bc879014 605 main/binary-amd64/Packages
 65dca66c72b18d59cdcf671775104e86cbe2123a 407 main/binary-amd64/Packages.gz
 c9810732c61aa7de2887b5194c6a09d0b6118664 79 main/binary-amd64/Release
 4cdb64c700f798f719f5c81ae42e44582be094c5 538 restricted/binary-i386/Packages
 190f980fd80d58284129ee050f9eb70b9590fedb 384 restricted/binary-i386/Packages.gz
 b1d1a4d57f5c8d70184c9661a087b8a92406c76d 84 restricted/binary-i386/Release
 4cdb64c700f798f719f5c81ae42e44582be094c5 538 restricted/binary-amd64/Packages
 190f980fd80d58284129ee050f9eb70b9590fedb 384 restricted/binary-amd64/Packages.gz
 4bd64fb2ef44037254729ab514d3403a65db7123 85 restricted/binary-amd64/Release
""",
            "hardy-updates": """Origin: Ubuntu
Label: Ubuntu
Codename: hardy-updates
Version: 8.04
Date: Tue, 31 Mar 2009 13:32:17 +0000
Architectures: i386 amd64
Components: main restricted
MD5Sum:
 a23ba734dc4fe7c1ec8dc960cc670b8e 1227 main/binary-i386/Packages
 2d6d271964be8000808abfa2b0e999b7 713 main/binary-i386/Packages.gz
 2f6ee66ed2d2b4115fabc8eed428e42e 78 main/binary-i386/Release
 a23ba734dc4fe7c1ec8dc960cc670b8e 1227 main/binary-amd64/Packages
 2d6d271964be8000808abfa2b0e999b7 713 main/binary-amd64/Packages.gz
 7e40db962fe49b6db232bf559cf6f79d 79 main/binary-amd64/Release
 d41d8cd98f00b204e9800998ecf8427e 0 restricted/binary-i386/Packages
 7029066c27ac6f5ef18d660d5741979a 20 restricted/binary-i386/Packages.gz
 8686ad9c5d83484dc66a1eca2bd8030f 84 restricted/binary-i386/Release
 d41d8cd98f00b204e9800998ecf8427e 0 restricted/binary-amd64/Packages
 7029066c27ac6f5ef18d660d5741979a 20 restricted/binary-amd64/Packages.gz
 6e24798a6089cd3a21226182784995e9 85 restricted/binary-amd64/Release
SHA1:
 9867c9f7ebbb5741fc589d0d4395ea8f74f3b5e4 1227 main/binary-i386/Packages
 2a7061fa162a607a63453c0360678052a38f0259 713 main/binary-i386/Packages.gz
 5fe86036c60d6210b662df3acc238e2936f03581 78 main/binary-i386/Release
 9867c9f7ebbb5741fc589d0d4395ea8f74f3b5e4 1227 main/binary-amd64/Packages
 2a7061fa162a607a63453c0360678052a38f0259 713 main/binary-amd64/Packages.gz
 c9810732c61aa7de2887b5194c6a09d0b6118664 79 main/binary-amd64/Release
 da39a3ee5e6b4b0d3255bfef95601890afd80709 0 restricted/binary-i386/Packages
 46c6643f07aa7f6bfe7118de926b86defc5087c4 20 restricted/binary-i386/Packages.gz
 b1d1a4d57f5c8d70184c9661a087b8a92406c76d 84 restricted/binary-i386/Release
 da39a3ee5e6b4b0d3255bfef95601890afd80709 0 restricted/binary-amd64/Packages
 46c6643f07aa7f6bfe7118de926b86defc5087c4 20 restricted/binary-amd64/Packages.gz
 4bd64fb2ef44037254729ab514d3403a65db7123 85 restricted/binary-amd64/Release
"""}

PACKAGES = {"hardy":
            {"restricted":
             {"amd64": """Package: kairos
Version: 0.0.8
Architecture: all
Maintainer: Free Ekanayaka <freee@debian.org>
Installed-Size: 192
Pre-Depends: libaugeas0, python-augeas, augeas-tools, jackd, rotter, monit, darkice, soma, python-remix, nfs-kernel-server, icecast2
Priority: extra
Section: admin
Filename: pool/restricted/k/kairos/kairos_0.0.8_all.deb
Size: 60768
SHA1: 1e5cc71cbd33d2b26a8feb19a48e815f271cd335
MD5sum: 5fd717ed3d15db25ffaa9d05fec62e42
Description: kairos customisation package
 This package configures and customises an kairos
 machine.

""",
              "i386": """Package: kairos
Version: 0.0.8
Architecture: all
Maintainer: Free Ekanayaka <freee@debian.org>
Installed-Size: 192
Pre-Depends: libaugeas0, python-augeas, augeas-tools, jackd, rotter, monit, darkice, soma, python-remix, nfs-kernel-server, icecast2
Priority: extra
Section: admin
Filename: pool/restricted/k/kairos/kairos_0.0.8_all.deb
Size: 60768
SHA1: 1e5cc71cbd33d2b26a8feb19a48e815f271cd335
MD5sum: 5fd717ed3d15db25ffaa9d05fec62e42
Description: kairos customisation package
 This package configures and customises an kairos
 machine.

"""},
             "main":
             {"amd64": """Package: libclthreads2
Source: clthreads
Version: 2.4.0-1
Architecture: amd64
Maintainer: Debian Multimedia Maintainers <pkg-multimedia-maintainers@lists.alioth.debian.org>
Installed-Size: 80
Depends: libc6 (>= 2.3.2), libgcc1 (>= 1:4.1.1), libstdc++6 (>= 4.1.1)
Priority: extra
Section: libs
Filename: pool/main/c/clthreads/libclthreads2_2.4.0-1_amd64.deb
Size: 12938
SHA1: dc6cb78896642dd436851888b8bd4454ab8f421b
MD5sum: 19960adb88e178fb7eb4997b47eee05b
Description: POSIX threads C++ access library
 C++ wrapper library around the POSIX threads API. This package includes
 the shared library object.

""",
              "i386": """Package: syslinux
Version: 2:3.73+dfsg-2
Architecture: i386
Maintainer: Daniel Baumann <daniel@debian.org>
Installed-Size: 140
Depends: libc6 (>= 2.7-1), syslinux-common (= 2:3.73+dfsg-2), dosfstools, mtools
Homepage: http://syslinux.zytor.com/
Priority: optional
Section: utils
Filename: pool/main/s/syslinux/syslinux_3.73+dfsg-2_i386.deb
Size: 70384
SHA1: 6edf6a7e81a5e9759270872e45c782394dfa85e5
MD5sum: ae8baa9f6c6a172a3b127af1e6675046
Description: utilities for the syslinux bootloaders
 SYSLINUX is a suite of lightweight bootloaders, currently supporting DOS FAT
 filesystems (SYSLINUX), Linux ext2/ext3 filesystems (EXTLINUX), PXE network
 booting (PXELINUX), or bootable "El Torito" ISO 9660 CD-ROMs (ISOLINUX). It
 also includes a tool, MEMDISK, which loads legacy operating systems (such as
 DOS) from these media.

"""}},
            "hardy-updates":
            {"restricted":
             {"amd64": """""",
              "i386": """"""},
             "main":
             {"amd64": """Package: rebuildd
Version: 0.3.5
Architecture: all
Maintainer: Julien Danjou <acid@debian.org>
Installed-Size: 312
Depends: python (>= 2.5), python-support (>= 0.7.1), lsb-base, python-sqlobject, python-apt
Recommends: pbuilder, python-gdchart2, python-webpy
Suggests: cowdancer
Priority: extra
Section: devel
Filename: pool/main/r/rebuildd/rebuildd_0.3.5_all.deb
Size: 24652
SHA1: 5446cd5c8a29212b403214884cae96f14824a573
MD5sum: 92e81240c2caf286ad103e44dcdc44e1
Description: build daemon aiming at rebuilding Debian packages
 This software allows you to manage a set of jobs. Each job is a package
 rebuilding task. Rebuilding is done by pbuilder (or cowbuilder if you want),
 or anything else, since everything is customizable via configuration file.
 It can also send build logs by email, event each log can be sent to a different
 email address.
 .
 rebuildd is multi-threaded, so you can run multiple build jobs in parallel.
 It is also administrable via a telnet interface. A Web interface is also
 embedded so you can see your jobs queue and watch log file in real-time in your
 browser.
 .
 rebuildd is designed to be run on multiple hosts even with different
 architecture set, and to parallelize the rebuild tasks.

""",
              "i386": """Package: rebuildd
Version: 0.3.5
Architecture: all
Maintainer: Julien Danjou <acid@debian.org>
Installed-Size: 312
Depends: python (>= 2.5), python-support (>= 0.7.1), lsb-base, python-sqlobject, python-apt
Recommends: pbuilder, python-gdchart2, python-webpy
Suggests: cowdancer
Priority: extra
Section: devel
Filename: pool/main/r/rebuildd/rebuildd_0.3.5_all.deb
Size: 24652
SHA1: 5446cd5c8a29212b403214884cae96f14824a573
MD5sum: 92e81240c2caf286ad103e44dcdc44e1
Description: build daemon aiming at rebuilding Debian packages
 This software allows you to manage a set of jobs. Each job is a package
 rebuilding task. Rebuilding is done by pbuilder (or cowbuilder if you want),
 or anything else, since everything is customizable via configuration file.
 It can also send build logs by email, event each log can be sent to a different
 email address.
 .
 rebuildd is multi-threaded, so you can run multiple build jobs in parallel.
 It is also administrable via a telnet interface. A Web interface is also
 embedded so you can see your jobs queue and watch log file in real-time in your
 browser.
 .
 rebuildd is designed to be run on multiple hosts even with different
 architecture set, and to parallelize the rebuild tasks.

"""}}}


def create_deb(target_dir, pkg_name, pkg_data):
    """Create a Debian package in the specified C{target_dir}."""
    path = os.path.join(target_dir, pkg_name)
    data = base64.decodestring(pkg_data)
    create_file(path, data)


def create_simple_repository(target_dir):
    """Create a simple deb-dir repository with in C{target_dir}."""
    create_deb(target_dir, PKGNAME1, PKGDEB1)
    create_deb(target_dir, PKGNAME2, PKGDEB2)
    create_deb(target_dir, PKGNAME3, PKGDEB3)


def create_full_repository(target_dir):
    """
    Create a full APT repository with a dists/ tree rooted at C{target_dir}.
    """

    class Repository(object):

        codename = "hardy"
        variant = "hardy-updates"
        components = ["main", "restricted"]
        archs = ["amd64", "i386"]
        hashes = [HASH4, HASH5, HASH6, HASH7]

        def __init__(self, root):
            self.root = root
            self.url = "file://%s" % self.root

    repository = Repository(target_dir)
    dists_directory = os.path.join(repository.root, "dists")
    os.mkdir(dists_directory)
    for dist in [repository.codename, repository.variant]:
        dist_directory = os.path.join(dists_directory, dist)
        os.mkdir(dist_directory)
        fd = open(os.path.join(dist_directory, "Release"), "w")
        fd.write(RELEASES[dist])
        fd.close()
        for component in repository.components:
            component_directory = os.path.join(dist_directory, component)
            os.mkdir(component_directory)
            for arch in repository.archs:
                arch_directory = os.path.join(component_directory,
                                              "binary-%s" % arch)
                os.mkdir(arch_directory)
                fd = open(os.path.join(arch_directory, "Packages"), "w")
                fd.write(PACKAGES[dist][component][arch])
                fd.close()
                fd = open(os.path.join(arch_directory, "Release"), "w")
                fd.write("""Version: 8.04
Component: %s
Origin: Ubuntu
Label: Ubuntu
Architecture: %s
""" % (component, arch))
                fd.close()
    return repository



import warnings

from twisted.trial.unittest import TestCase

from landscape.lib.warning import hide_warnings


class WarningTest(TestCase):
    def setUp(self):
        super(WarningTest, self).setUp()
        self.orig_filters = warnings.filters[:]

    def tearDown(self):
        super(WarningTest, self).tearDown()
        warnings.filters[:] = self.orig_filters

    def test_hide_warnings(self):
        hide_warnings()
        filters = warnings.filters[:2]

        # Warning filters are processed beginning to end, and the first filter
        # which matches a particular warning is used.

        self.assertEqual(
            filters,
            # The frontmost should "default" (i.e.  print) on UserWarnings
            [("default", None, UserWarning, None, 0),
             # The one just behind that should indicate that we should ignore
             # all other warnings.
             ("ignore", None, Warning, None, 0)])



import sys

from landscape.ui.tests.helpers import (
    ConfigurationProxyHelper, dbus_test_should_skip, dbus_skip_message,
    FakeGSettings)

if not dbus_test_should_skip:
    from gi.repository import Gtk
    from landscape.ui.controller.app import SettingsApplicationController
    from landscape.ui.controller.configuration import ConfigController
    from landscape.ui.view.configuration import ClientSettingsDialog
    from landscape.ui.model.configuration.uisettings import UISettings
else:
    SettingsApplicationController = object

from landscape.tests.helpers import LandscapeTest


class ConnectionRecordingSettingsApplicationController(
    SettingsApplicationController):
    _connections = set()
    _connection_args = {}
    _connection_kwargs = {}

    def __init__(self, get_config=None, get_uisettings=None):
        super(ConnectionRecordingSettingsApplicationController,
              self).__init__()
        if get_config:
            self.get_config = get_config
        if get_uisettings:
            self.get_uisettings = get_uisettings

    def _make_connection_name(self, signal, func):
        return signal + ">" + func.__name__

    def _record_connection(self, signal, func, *args, **kwargs):
        connection_name = self._make_connection_name(signal, func)
        self._connections.add(connection_name)
        signal_connection_args = self._connection_args.get(
            connection_name, [])
        signal_connection_args.append(repr(args))
        self._connection_args = signal_connection_args
        signal_connection_kwargs = self._connection_kwargs.get(
            connection_name, [])
        signal_connection_kwargs.append(repr(kwargs))
        self._connection_kwargs = signal_connection_kwargs

    def is_connected(self, signal, func):
        connection_name = self._make_connection_name(signal, func)
        return self._connections.issuperset(set([connection_name]))

    def connect(self, signal, func, *args, **kwargs):
        self._record_connection(signal, func)


class SettingsApplicationControllerInitTest(LandscapeTest):

    def setUp(self):
        super(SettingsApplicationControllerInitTest, self).setUp()

    def test_init(self):
        """
        Test we connect activate to something useful on application
        initialisation.
        """
        app = ConnectionRecordingSettingsApplicationController()
        self.assertTrue(app.is_connected("activate", app.setup_ui))

    if dbus_test_should_skip:
        skip = dbus_skip_message


class SettingsApplicationControllerUISetupTest(LandscapeTest):

    helpers = [ConfigurationProxyHelper]

    def setUp(self):
        self.config_string = "\n".join(
            ["[client]",
             "data_path = %s" % sys.path[0],
             "http_proxy = http://proxy.localdomain:3192",
             "tags = a_tag",
             "url = https://landscape.canonical.com/message-system",
             "account_name = foo",
             "registration_key = bar",
             "computer_title = baz",
             "https_proxy = https://proxy.localdomain:6192",
             "ping_url = http://landscape.canonical.com/ping"])
        self.default_data = {"management-type": "not",
                             "computer-title": "",
                             "hosted-landscape-host": "",
                             "hosted-account-name": "",
                             "hosted-password": "",
                             "local-landscape-host": "",
                             "local-account-name": "",
                             "local-password": ""}
        super(SettingsApplicationControllerUISetupTest, self).setUp()

        def fake_run(obj):
            """
            Retard X11 mapping.
            """
            pass

        self._real_run = Gtk.Dialog.run
        Gtk.Dialog.run = fake_run

        def get_config():
            return self.proxy

        def get_uisettings():
            settings = FakeGSettings(data=self.default_data)
            return UISettings(settings)

        self.app = ConnectionRecordingSettingsApplicationController(
            get_config=get_config, get_uisettings=get_uisettings)

    def tearDown(self):
        Gtk.Dialog.run = self._real_run
        super(
            SettingsApplicationControllerUISetupTest, self).tearDown()

    def test_setup_ui(self):
        """
        Test that we correctly setup the L{ClientSettingsDialog} with
        the config object and correct data
        """
        self.assertRaises(SystemExit, self.app.setup_ui, data=None,
                          asynchronous=False)
        self.assertIsInstance(self.app.settings_dialog, ClientSettingsDialog)
        self.assertIsInstance(self.app.settings_dialog.controller,
                              ConfigController)

    if dbus_test_should_skip:
        skip = dbus_skip_message



# -*- coding: utf-8 -*-

from landscape.tests.helpers import LandscapeTest
from landscape.lib.encoding import encode_if_needed, encode_dict_if_needed


class EncodingTest(LandscapeTest):

    def test_encode_if_needed_utf_string(self):
        """
        When passed an utf-8 str() instance the encode_if_needed function
        returns the same.
        """
        value = "請不要刪除"
        result = encode_if_needed(value)
        self.assertEqual(value, result)

    def test_encode_if_needed_utf16_string(self):
        """
        When passed an unicode instance that is a decode()'d unicode (utf-16),
        the encode_if_needed function returns the utf-16 str() equivalent
        (in utf-8).
        """
        value = u"Alex \U0001f603"
        result = encode_if_needed(value)
        expected = 'Alex \xf0\x9f\x98\x83'
        self.assertEqual(expected, result)

    def test_encode_if_needed_utf_unicode(self):
        """
        When passed an unicode instance that is a decode()'d unicode,
        the encode_if_needed function returns the utf-8 str() equivalent.
        """
        value = u'\u8acb\u4e0d\u8981\u522a\u9664'
        expected = "請不要刪除"
        result = encode_if_needed(value)
        self.assertEqual(expected, result)

    def test_encode_if_needed_utf_unicode_string(self):
        """
        When passed an encoded() unicode instance, the encode_if_needed
        function returns the utf-8 str() equivalent.
        """
        value = u"請不要刪除"
        expected = "請不要刪除"
        result = encode_if_needed(value)
        self.assertEqual(expected, result)

    def test_encode_if_needed_with_null_value(self):
        """
        When passed None, the encode_if_needed function returns None.
        """
        self.assertIs(None, encode_if_needed(None))

    def test_encode_dict_if_needed(self):
        """
        The encode_dict_if_needed function returns a dict for which every
        value was passed to the encode_if_needed function.
        """
        value = {"a": "請不要刪除", "b": u'\u8acb\u4e0d\u8981\u522a\u9664',
                 "c": u"請不要刪除", "d": None, "e": 123}
        expected = {"a": "請不要刪除", "b": "請不要刪除", "c": "請不要刪除",
                    "d": None, "e": 123}
        result = encode_dict_if_needed(value)
        self.assertEqual(expected, result)



"""
Functionality for running arbitrary shell scripts.

@var ALL_USERS: A token indicating all users should be allowed.
"""
import os
import pwd
import tempfile
import operator
import shutil

from twisted.internet.protocol import ProcessProtocol
from twisted.internet.defer import (
    Deferred, fail, inlineCallbacks, returnValue, succeed)
from twisted.internet.error import ProcessDone

from landscape import VERSION
from landscape.constants import UBUNTU_PATH
from landscape.lib.scriptcontent import build_script
from landscape.lib.fetch import fetch_async, HTTPCodeError
from landscape.lib.persist import Persist
from landscape.lib.encoding import encode_if_needed
from landscape.manager.plugin import ManagerPlugin, SUCCEEDED, FAILED


ALL_USERS = object()
TIMEOUT_RESULT = 102
PROCESS_FAILED_RESULT = 103
FETCH_ATTACHMENTS_FAILED_RESULT = 104


class UnknownUserError(Exception):
    pass


def get_user_info(username=None):
    uid = None
    gid = None
    path = None
    if username is not None:
        username_str = encode_if_needed(username)
        try:
            info = pwd.getpwnam(username_str)
        except KeyError:
            raise UnknownUserError(u"Unknown user '%s'" % username)
        uid = info.pw_uid
        gid = info.pw_gid
        path = info.pw_dir
        if not os.path.exists(path):
            path = "/"
    return (uid, gid, path)


class ProcessTimeLimitReachedError(Exception):
    """
    Raised when a process has been running for too long.

    @ivar data: The data that the process printed before reaching the time
        limit.
    """

    def __init__(self, data):
        self.data = data


class ProcessFailedError(Exception):
    """Raised when a process exits with a non-0 exit code.

    @ivar data: The data that the process printed before reaching the time
        limit.
    """

    def __init__(self, data, exit_code):
        self.data = data
        self.exit_code = exit_code


class UnknownInterpreterError(Exception):
    """Raised when the interpreter specified to run a script is invalid.

    @ivar interpreter: the interpreter specified for the script.
    """

    def __init__(self, interpreter):
        self.interpreter = interpreter
        Exception.__init__(self, self._get_message())

    def _get_message(self):
        return "Unknown interpreter: '%s'" % self.interpreter


class ScriptRunnerMixin(object):
    """
    @param process_factory: The L{IReactorProcess} provider to run the
        process with.
    """

    def __init__(self, process_factory=None):
        if process_factory is None:
            from twisted.internet import reactor as process_factory
        self.process_factory = process_factory

    def is_user_allowed(self, user):
        allowed_users = self.registry.config.get_allowed_script_users()
        return allowed_users == ALL_USERS or user in allowed_users

    def write_script_file(self, script_file, filename, shell, code, uid, gid):
        # Chown and chmod it before we write the data in it - the script may
        # have sensitive content
        # It would be nice to use fchown(2) and fchmod(2), but they're not
        # available in python and using it with ctypes is pretty tedious, not
        # to mention we can't get errno.
        os.chmod(filename, 0700)
        if uid is not None:
            os.chown(filename, uid, gid)
        script_file.write(build_script(shell, code))
        script_file.close()

    def _run_script(self, filename, uid, gid, path, env, time_limit):

        if uid == os.getuid():
            uid = None
        if gid == os.getgid():
            gid = None

        pp = ProcessAccumulationProtocol(
            self.registry.reactor, self.size_limit)
        self.process_factory.spawnProcess(
            pp, filename, uid=uid, gid=gid, path=path, env=env)
        if time_limit is not None:
            pp.schedule_cancel(time_limit)
        return pp.result_deferred


class ScriptExecutionPlugin(ManagerPlugin, ScriptRunnerMixin):
    """A plugin which allows execution of arbitrary shell scripts.

    @ivar size_limit: The number of bytes at which to truncate process output.
    """

    size_limit = 500000

    def register(self, registry):
        super(ScriptExecutionPlugin, self).register(registry)
        registry.register_message(
            "execute-script", self._handle_execute_script)

    def _respond(self, status, data, opid, result_code=None):
        if not isinstance(data, unicode):
            # Let's decode result-text, replacing non-printable
            # characters
            data = data.decode("utf-8", "replace")
        message = {"type": "operation-result",
                   "status": status,
                   "result-text": data,
                   "operation-id": opid}
        if result_code:
            message["result-code"] = result_code
        return self.registry.broker.send_message(
            message, self._session_id, True)

    def _handle_execute_script(self, message):
        opid = message["operation-id"]
        try:
            user = message["username"]
            if not self.is_user_allowed(user):
                return self._respond(
                    FAILED,
                    u"Scripts cannot be run as user %s." % (user,),
                    opid)
            server_supplied_env = message.get("env", None)

            d = self.run_script(message["interpreter"], message["code"],
                                time_limit=message["time-limit"], user=user,
                                attachments=message["attachments"],
                                server_supplied_env=server_supplied_env)
            d.addCallback(self._respond_success, opid)
            d.addErrback(self._respond_failure, opid)
            return d
        except Exception, e:
            self._respond(FAILED, self._format_exception(e), opid)
            raise

    def _format_exception(self, e):
        return u"%s: %s" % (e.__class__.__name__, e.args[0])

    def _respond_success(self, data, opid):
        return self._respond(SUCCEEDED, data, opid)

    def _respond_failure(self, failure, opid):
        code = None
        if failure.check(ProcessTimeLimitReachedError):
            code = TIMEOUT_RESULT
        elif failure.check(ProcessFailedError):
            code = PROCESS_FAILED_RESULT
        elif failure.check(HTTPCodeError):
            code = FETCH_ATTACHMENTS_FAILED_RESULT
            return self._respond(
                FAILED, str(failure.value), opid,
                FETCH_ATTACHMENTS_FAILED_RESULT)

        if code is not None:
            return self._respond(FAILED, failure.value.data, opid, code)
        else:
            return self._respond(FAILED, str(failure), opid)

    @inlineCallbacks
    def _save_attachments(self, attachments, uid, gid, computer_id):
        root_path = self.registry.config.url.rsplit("/", 1)[0] + "/attachment/"
        attachment_dir = tempfile.mkdtemp()
        headers = {"User-Agent": "landscape-client/%s" % VERSION,
                   "Content-Type": "application/octet-stream",
                   "X-Computer-ID": computer_id}
        for filename, attachment_id in attachments.items():
            if isinstance(attachment_id, str):
                # Backward compatible behavior
                data = attachment_id
                yield succeed(None)
            else:
                data = yield fetch_async(
                    "%s%d" % (root_path, attachment_id),
                    cainfo=self.registry.config.ssl_public_key,
                    headers=headers)
            full_filename = os.path.join(attachment_dir, filename)
            attachment = file(full_filename, "wb")
            os.chmod(full_filename, 0600)
            if uid is not None:
                os.chown(full_filename, uid, gid)
            attachment.write(data)
            attachment.close()
        os.chmod(attachment_dir, 0700)
        if uid is not None:
            os.chown(attachment_dir, uid, gid)
        returnValue(attachment_dir)

    def run_script(self, shell, code, user=None, time_limit=None,
                   attachments=None, server_supplied_env=None):
        """
        Run a script based on a shell and the code.

        A file will be written with #!<shell> as the first line, as executable,
        and run as the given user.

        XXX: Handle the 'reboot' and 'killall landscape-client' commands
        gracefully.

        @param shell: The interpreter to use.
        @param code: The code to run.
        @param user: The username to run the process as.
        @param time_limit: The number of seconds to allow the process to run
            before killing it and failing the returned Deferred with a
            L{ProcessTimeLimitReachedError}.
        @param attachments: C{dict} of filename/data attached to the script.

        @return: A deferred that will fire with the data printed by the process
            or fail with a L{ProcessTimeLimitReachedError}.
        """
        if not os.path.exists(shell.split()[0]):
            return fail(
                UnknownInterpreterError(shell))

        uid, gid, path = get_user_info(user)

        fd, filename = tempfile.mkstemp()
        script_file = os.fdopen(fd, "w")
        self.write_script_file(
            script_file, filename, shell, code, uid, gid)

        env = {"PATH": UBUNTU_PATH, "USER": user or "", "HOME": path or ""}
        if server_supplied_env:
            env.update(server_supplied_env)
        old_umask = os.umask(0022)

        if attachments:
            persist = Persist(
                filename=os.path.join(self.registry.config.data_path,
                                      "broker.bpickle"))
            persist = persist.root_at("registration")
            computer_id = persist.get("secure-id")
            d = self._save_attachments(attachments, uid, gid, computer_id)
        else:
            d = succeed(None)

        def prepare_script(attachment_dir):

            if attachment_dir is not None:
                env["LANDSCAPE_ATTACHMENTS"] = attachment_dir

            return self._run_script(
                filename, uid, gid, path, env, time_limit)

        d.addCallback(prepare_script)
        return d.addBoth(self._cleanup, filename, env, old_umask)

    def _cleanup(self, result, filename, env, old_umask):
        try:
            os.unlink(filename)
        except:
            pass
        if "LANDSCAPE_ATTACHMENTS" in env:
            try:
                shutil.rmtree(env["LANDSCAPE_ATTACHMENTS"])
            except:
                pass
        os.umask(old_umask)
        return result


class ProcessAccumulationProtocol(ProcessProtocol):
    """A ProcessProtocol which accumulates output.

    @ivar size_limit: The number of bytes at which to truncate output.
    """

    def __init__(self, reactor, size_limit):
        self.data = []
        self.result_deferred = Deferred()
        self._cancelled = False
        self.size_limit = size_limit
        self.reactor = reactor
        self._scheduled_cancel = None

    def schedule_cancel(self, time_limit):
        self._scheduled_cancel = self.reactor.call_later(
            time_limit, self._cancel)

    def childDataReceived(self, fd, data):
        """Some data was received from the child.

        Add it to our buffer, as long as it doesn't go over L{size_limit}
        bytes.
        """
        current_size = reduce(operator.add, map(len, self.data), 0)
        self.data.append(data[:self.size_limit - current_size])

    def processEnded(self, reason):
        """Fire back the deferred.

        The deferred will be fired with the string of data received from the
        subprocess, or if the subprocess was cancelled, a
        L{ProcessTimeLimitReachedError} will be fired with data accumulated so
        far.
        """
        exit_code = reason.value.exitCode
        data = "".join(self.data)
        if self._cancelled:
            self.result_deferred.errback(ProcessTimeLimitReachedError(data))
        else:
            if self._scheduled_cancel is not None:
                scheduled = self._scheduled_cancel
                self._scheduled_cancel = None
                self.reactor.cancel_call(scheduled)

            if reason.check(ProcessDone):
                self.result_deferred.callback(data)
            else:
                self.result_deferred.errback(ProcessFailedError(data,
                                                                exit_code))

    def _cancel(self):
        """
        Close filedescriptors, kill the process, and indicate that a
        L{ProcessTimeLimitReachedError} should be fired on the deferred.
        """
        # Sometimes children of the shell we're killing won't die unless their
        # file descriptors are closed! For example, if /bin/sh -c "cat" is the
        # process, "cat" won't die when we kill its shell. I'm not sure if this
        # is really sufficient: maybe there's a way we can walk over all
        # children of the process we started and kill them all.
        for i in (0, 1, 2):
            self.transport.closeChildFD(i)
        self.transport.signalProcess("KILL")
        self._cancelled = True


class ScriptExecution(ManagerPlugin):
    """
    Meta-plugin wrapping ScriptExecutionPlugin and CustomGraphPlugin.
    """

    def __init__(self):
        from landscape.manager.customgraph import CustomGraphPlugin
        self._script_execution = ScriptExecutionPlugin()
        self._custom_graph = CustomGraphPlugin()

    def register(self, registry):
        super(ScriptExecution, self).register(registry)
        self._script_execution.register(registry)
        self._custom_graph.register(registry)

    def exchange(self, urgent=False):
        self._custom_graph.exchange(urgent)



from landscape.deployment import Configuration


ALL_PLUGINS = ["ActiveProcessInfo", "ComputerInfo",
               "LoadAverage", "MemoryInfo", "MountInfo", "ProcessorInfo",
               "Temperature", "PackageMonitor", "UserMonitor",
               "RebootRequired", "AptPreferences", "NetworkActivity",
               "NetworkDevice", "UpdateManager", "CPUUsage", "SwiftUsage",
               "CephUsage"]


class MonitorConfiguration(Configuration):
    """Specialized configuration for the Landscape Monitor."""

    def make_parser(self):
        """
        Specialize L{Configuration.make_parser}, adding many
        monitor-specific options.
        """
        parser = super(MonitorConfiguration, self).make_parser()

        parser.add_option("--monitor-plugins", metavar="PLUGIN_LIST",
                          help="Comma-delimited list of monitor plugins to "
                               "use. ALL means use all plugins.",
                          default="ALL")
        return parser

    @property
    def plugin_factories(self):
        if self.monitor_plugins == "ALL":
            return ALL_PLUGINS
        return [x.strip() for x in self.monitor_plugins.split(",")]



import tempfile
from landscape.lib.fs import touch_file
from landscape.tests.helpers import LandscapeTest, MonitorHelper
from landscape.monitor.cephusage import CephUsage


class CephUsagePluginTest(LandscapeTest):
    helpers = [MonitorHelper]

    def setUp(self):
        super(CephUsagePluginTest, self).setUp()
        self.mstore = self.broker_service.message_store
        self.plugin = CephUsage(create_time=self.reactor.time)

    def test_never_exchange_empty_messages(self):
        """
        The plugin will create a message with an empty C{data-points} list
        when no previous data is available.  If an empty message is created
        during exchange, it should not be queued.
        """
        self.mstore.set_accepted_types(["ceph-usage"])
        self.monitor.add(self.plugin)

        self.monitor.exchange()
        self.assertEqual(0, len(self.mstore.get_pending_messages()))

    def test_exchange_messages(self):
        """
        The Ceph usage plugin queues message when manager.exchange()
        is called.
        """
        ring_id = "whatever"
        self.mstore.set_accepted_types(["ceph-usage"])

        point = (60, 100000, 80000, 20000)
        self.plugin._ceph_usage_points = [point]
        self.plugin._ceph_ring_id = ring_id
        self.monitor.add(self.plugin)

        self.monitor.exchange()
        self.assertMessages(
            self.mstore.get_pending_messages(),
            [{"type": "ceph-usage",
              "ring-id": ring_id,
              "ceph-usages": [],
              "data-points": [point]}])

    def test_create_message(self):
        """
        Calling create_message returns an expected message.
        """
        ring_id = "blah"
        self.plugin._ceph_usage_points = []
        self.plugin._ceph_ring_id = ring_id
        message = self.plugin.create_message()

        self.assertIn("type", message)
        self.assertEqual(message["type"], "ceph-usage")
        self.assertIn("data-points", message)
        self.assertEqual(ring_id, message["ring-id"])
        data_points = message["data-points"]
        self.assertEqual(len(data_points), 0)

        point = (60, 100000, 80000, 20000)
        self.plugin._ceph_usage_points = [point]
        message = self.plugin.create_message()
        self.assertIn("type", message)
        self.assertEqual(message["type"], "ceph-usage")
        self.assertIn("data-points", message)
        self.assertEqual(ring_id, message["ring-id"])
        data_points = message["data-points"]
        self.assertEqual(len(data_points), 1)
        self.assertEqual([point], data_points)

    def test_no_message_if_not_accepted(self):
        """
        Don't add any messages at all if the broker isn't currently accepting
        their type.
        """
        interval = 30
        monitor_interval = 300

        plugin = CephUsage(
            interval=interval, monitor_interval=monitor_interval,
            create_time=self.reactor.time)

        self.monitor.add(plugin)

        self.reactor.advance(monitor_interval * 2)
        self.monitor.exchange()

        self.mstore.set_accepted_types(["ceph-usage"])
        self.assertMessages(list(self.mstore.get_pending_messages()), [])

    def test_wb_should_run_inactive(self):
        """
        A plugin with self.active set to False should not run.
        """
        plugin = CephUsage()
        plugin.active = False
        self.assertFalse(plugin._should_run())

    def test_wb_should_run_no_config_file(self):
        """
        A plugin without a set _ceph_config attribute should not run.
        """
        plugin = CephUsage()
        plugin._has_rados = True
        plugin._ceph_config = None
        self.assertFalse(plugin._should_run())

    def test_wb_should_run_no_rados(self):
        """
        If the Rados library cannot be imported (CephUsage._has_rados is False)
        the plugin logs a message then deactivates itself.
        """
        logging_mock = self.mocker.replace("logging.info")
        logging_mock("This machine does not appear to be a Ceph machine. "
                     "Deactivating plugin.")
        self.mocker.replay()

        plugin = CephUsage()
        plugin._has_rados = False
        self.assertFalse(plugin._should_run())

    def test_wb_should_run(self):
        """
        If the Rados library is present with the correct version and a ceph
        config exists, the C{_should_run} method returns True.
        """
        plugin = CephUsage()
        plugin._has_rados = True
        plugin._ceph_config = tempfile.mktemp()
        touch_file(plugin._ceph_config)
        self.assertTrue(plugin._should_run())

    def test_wb_handle_usage(self):
        """
        The C{_handle_usage} method stores the result of the rados call (here,
        an example value) in an Accumulator, and appends the step_data
        to the C{_ceph_usage_points} member when an accumulator interval is
        reached.
        """
        interval = 300
        stats = {"kb": 10240L, "kb_avail": 8192L, "kb_used": 2048L}

        plugin = CephUsage(
            create_time=self.reactor.time, interval=interval,
            monitor_interval=interval)

        self.monitor.add(plugin)

        plugin._handle_usage(stats)  # time is 0

        self.reactor.advance(interval)  # time is 300
        plugin._handle_usage(stats)

        self.assertEqual(
            [(300, 10485760, 8388608, 2097152)], plugin._ceph_usage_points)

    def test_resynchronize_message_calls_reset_method(self):
        """
        If the reactor fires a "resynchronize" even the C{_reset}
        method on the ceph plugin object is called.
        """
        self.called = False

        def stub_reset():
            self.called = True

        self.plugin._reset = stub_reset
        self.monitor.add(self.plugin)
        self.reactor.fire("resynchronize")
        self.assertTrue(self.called)



import os

from twisted.internet.defer import succeed


class Load(object):

    def register(self, sysinfo):
        self._sysinfo = sysinfo

    def run(self):
        self._sysinfo.add_header("System load", str(os.getloadavg()[0]))
        return succeed(None)



# -*- encoding: utf-8 -*-
import base64
import time
import sys
import os

from twisted.internet.defer import Deferred
from twisted.python.failure import Failure
from twisted.internet.error import ProcessTerminated, ProcessDone

from landscape.lib.fs import create_file, read_file, touch_file
from landscape.package.changer import (
    PackageChanger, main, find_changer_command, UNKNOWN_PACKAGE_DATA_TIMEOUT,
    SUCCESS_RESULT, DEPENDENCY_ERROR_RESULT, POLICY_ALLOW_INSTALLS,
    POLICY_ALLOW_ALL_CHANGES, ERROR_RESULT)
from landscape.package.store import PackageStore
from landscape.package.facade import (
    DependencyError, TransactionError)
from landscape.package.changer import (
    PackageChangerConfiguration, ChangePackagesResult)
from landscape.tests.mocker import ANY
from landscape.tests.helpers import (
    LandscapeTest, BrokerServiceHelper, StubProcessFactory)
from landscape.package.tests.helpers import (
    HASH1, HASH2, HASH3, PKGDEB1, PKGDEB2,
    AptFacadeHelper, SimpleRepositoryHelper)
from landscape.manager.manager import FAILED
from landscape.manager.shutdownmanager import ShutdownFailedError
from landscape.reactor import FakeReactor


class AptPackageChangerTest(LandscapeTest):

    helpers = [AptFacadeHelper, SimpleRepositoryHelper, BrokerServiceHelper]

    def setUp(self):
        super(AptPackageChangerTest, self).setUp()
        self.store = PackageStore(self.makeFile())
        self.config = PackageChangerConfiguration()
        self.config.data_path = self.makeDir()
        self.process_factory = StubProcessFactory()
        self.landscape_reactor = FakeReactor()
        reboot_required_filename = self.makeFile("reboot required")
        os.mkdir(self.config.package_directory)
        os.mkdir(self.config.binaries_path)
        touch_file(self.config.update_stamp_filename)
        self.changer = PackageChanger(
            self.store, self.facade, self.remote, self.config,
            process_factory=self.process_factory,
            landscape_reactor=self.landscape_reactor,
            reboot_required_filename=reboot_required_filename)
        self.changer.update_notifier_stamp = "/Not/Existing"
        self.changer.get_session_id()
        service = self.broker_service
        service.message_store.set_accepted_types(["change-packages-result",
                                                  "operation-result"])

    def set_pkg1_installed(self):
        """Return the hash of a package that is installed."""
        self._add_system_package("foo")
        self.facade.reload_channels()
        [foo] = self.facade.get_packages_by_name("foo")
        return self.facade.get_package_hash(foo)

    def set_pkg2_satisfied(self):
        """Return the hash of a package that can be installed."""
        self._add_package_to_deb_dir(self.repository_dir, "bar")
        self.facade.reload_channels()
        [bar] = self.facade.get_packages_by_name("bar")
        return self.facade.get_package_hash(bar)

    def set_pkg1_and_pkg2_satisfied(self):
        """Make a package depend on another package.

        Return the hashes of the two packages.
        """
        self._add_package_to_deb_dir(
            self.repository_dir, "foo", control_fields={"Depends": "bar"})
        self._add_package_to_deb_dir(self.repository_dir, "bar")
        self.facade.reload_channels()
        [foo] = self.facade.get_packages_by_name("foo")
        [bar] = self.facade.get_packages_by_name("bar")
        return (
            self.facade.get_package_hash(foo),
            self.facade.get_package_hash(bar))

    def set_pkg2_upgrades_pkg1(self):
        """Make it so that one package upgrades another.

        Return the hashes of the two packages.
        """
        self._add_system_package("foo", version="1.0")
        self._add_package_to_deb_dir(self.repository_dir, "foo", version="2.0")
        self.facade.reload_channels()
        foo_1, foo_2 = sorted(self.facade.get_packages_by_name("foo"))
        return (
            self.facade.get_package_hash(foo_1),
            self.facade.get_package_hash(foo_2))

    def remove_pkg2(self):
        """Remove package name2 from its repository."""
        packages_file = os.path.join(self.repository_dir, "Packages")
        packages_contents = read_file(packages_file)
        packages_contents = "\n\n".join(
            [stanza for stanza in packages_contents.split("\n\n")
             if "Package: name2" not in stanza])
        create_file(packages_file, packages_contents)

    def get_binaries_channels(self, binaries_path):
        """Return the channels that will be used for the binaries."""
        return [{"baseurl": "file://%s" % binaries_path,
                 "components": "",
                 "distribution": "./",
                 "type": "deb"}]

    def get_package_name(self, version):
        """Return the name of the package."""
        return version.package.name

    def disable_clear_channels(self):
        """Disable clear_channels(), so that it doesn't remove test setup.

        This is useful for change-packages tests, which will call
        facade.clear_channels(). Normally that's safe, but since we used
        the facade to set up channels, we don't want them to be removed.
        """
        self.facade.clear_channels = lambda: None

    def get_pending_messages(self):
        return self.broker_service.message_store.get_pending_messages()

    def replace_perform_changes(self, func):
        old_perform_changes = self.Facade.perform_changes

        def reset_perform_changes(Facade):
            Facade.perform_changes = old_perform_changes

        self.addCleanup(reset_perform_changes, self.Facade)
        self.Facade.perform_changes = func

    def test_unknown_package_id_for_dependency(self):
        hash1, hash2 = self.set_pkg1_and_pkg2_satisfied()

        # Let's request an operation that would require an answer with a
        # must-install field with a package for which the id isn't yet
        # known by the client.
        self.store.add_task("changer",
                            {"type": "change-packages", "install": [1],
                             "operation-id": 123})

        # In our first try, we should get nothing, because the id of the
        # dependency (hash2) isn't known.
        self.store.set_hash_ids({hash1: 1})
        result = self.changer.handle_tasks()
        self.assertEqual(result.called, True)
        self.assertMessages(self.get_pending_messages(), [])

        self.assertIn("Package data not yet synchronized with server (%r)"
                      % hash2, self.logfile.getvalue())

        # So now we'll set it, and advance the reactor to the scheduled
        # change detection.  We'll get a lot of messages, including the
        # result of our previous message, which got *postponed*.
        self.store.set_hash_ids({hash2: 2})
        result = self.changer.handle_tasks()

        def got_result(result):
            self.assertMessages(self.get_pending_messages(),
                                [{"must-install": [2],
                                  "operation-id": 123,
                                  "result-code": 101,
                                  "type": "change-packages-result"}])
        return result.addCallback(got_result)

    def test_install_unknown_id(self):
        self.store.add_task("changer",
                            {"type": "change-packages", "install": [456],
                             "operation-id": 123})

        self.changer.handle_tasks()

        self.assertIn("Package data not yet synchronized with server (456)",
                      self.logfile.getvalue())
        self.assertTrue(self.store.get_next_task("changer"))

    def test_remove_unknown_id(self):
        self.store.add_task("changer",
                            {"type": "change-packages", "remove": [456],
                             "operation-id": 123})

        self.changer.handle_tasks()

        self.assertIn("Package data not yet synchronized with server (456)",
                      self.logfile.getvalue())
        self.assertTrue(self.store.get_next_task("changer"))

    def test_install_unknown_package(self):
        self.store.set_hash_ids({"hash": 456})
        self.store.add_task("changer",
                            {"type": "change-packages", "install": [456],
                             "operation-id": 123})

        self.changer.handle_tasks()

        self.assertIn("Package data not yet synchronized with server ('hash')",
                      self.logfile.getvalue())
        self.assertTrue(self.store.get_next_task("changer"))

    def test_remove_unknown_package(self):
        self.store.set_hash_ids({"hash": 456})
        self.store.add_task("changer",
                            {"type": "change-packages", "remove": [456],
                             "operation-id": 123})

        self.changer.handle_tasks()

        self.assertIn("Package data not yet synchronized with server ('hash')",
                      self.logfile.getvalue())
        self.assertTrue(self.store.get_next_task("changer"))

    def test_unknown_data_timeout(self):
        """After a while, unknown package data is reported as an error.

        In these cases a warning is logged, and the task is removed.
        """
        self.store.add_task("changer",
                            {"type": "change-packages", "remove": [123],
                             "operation-id": 123})

        time_mock = self.mocker.replace("time.time")
        time_mock()
        self.mocker.result(time.time() + UNKNOWN_PACKAGE_DATA_TIMEOUT)
        self.mocker.count(1, None)
        self.mocker.replay()

        try:
            result = self.changer.handle_tasks()
            self.mocker.verify()
        finally:
            # Reset it earlier so that Twisted has the true time function.
            self.mocker.reset()

        self.assertIn("Package data not yet synchronized with server (123)",
                      self.logfile.getvalue())

        def got_result(result):
            message = {"type": "change-packages-result",
                       "operation-id": 123,
                       "result-code": 100,
                       "result-text": "Package data has changed. "
                                      "Please retry the operation."}
            self.assertMessages(self.get_pending_messages(), [message])
            self.assertEqual(self.store.get_next_task("changer"), None)
        return result.addCallback(got_result)

    def test_dependency_error(self):
        """
        In this test we hack the facade to simulate the situation where
        Smart didn't accept to remove the package due to missing
        dependencies that are present in the system but weren't requested
        in the message.

        The client must answer it saying which additional changes are
        needed to perform the requested operation.

        It's a slightly hackish approach, since we're returning
        the full set of packages available as a dependency error, but
        it serves well for testing this specific feature.
        """
        installed_hash = self.set_pkg1_installed()
        # Use ensure_channels_reloaded() to make sure that the package
        # instances we raise below are the same that the facade will
        # use. The changer will use ensure_channels_reloaded() too,
        # which won't actually reload the package data if it's already
        # loaded.
        self.facade.ensure_channels_reloaded()
        self.store.set_hash_ids({installed_hash: 1, HASH2: 2, HASH3: 3})
        self.store.add_task("changer",
                            {"type": "change-packages", "install": [2],
                             "operation-id": 123})

        packages = [
            self.facade.get_package_by_hash(pkg_hash)
            for pkg_hash in [installed_hash, HASH2, HASH3]]

        def raise_dependency_error(self):
            raise DependencyError(set(packages))

        self.replace_perform_changes(raise_dependency_error)

        result = self.changer.handle_tasks()

        def got_result(result):
            self.assertMessages(self.get_pending_messages(),
                                [{"must-install": [2, 3],
                                  "must-remove": [1],
                                  "operation-id": 123,
                                  "result-code": 101,
                                  "type": "change-packages-result"}])
        return result.addCallback(got_result)

    def test_dependency_error_with_binaries(self):
        """
        Simulate a failing operation involving server-generated binary
        packages. The extra changes needed to perform the transaction
        are sent back to the server.
        """
        self.remove_pkg2()
        installed_hash = self.set_pkg1_installed()
        self.store.set_hash_ids({installed_hash: 1, HASH3: 3})
        self.store.add_task("changer",
                            {"type": "change-packages",
                             "install": [2],
                             "binaries": [(HASH2, 2, PKGDEB2)],
                             "operation-id": 123})

        packages = set()

        def raise_dependency_error(self):
            packages.update(
                self.get_package_by_hash(pkg_hash)
                for pkg_hash in [installed_hash, HASH2, HASH3])
            raise DependencyError(set(packages))

        self.replace_perform_changes(raise_dependency_error)
        self.disable_clear_channels()

        result = self.changer.handle_tasks()

        def got_result(result):
            self.assertMessages(self.get_pending_messages(),
                                [{"must-install": [2, 3],
                                  "must-remove": [1],
                                  "operation-id": 123,
                                  "result-code": 101,
                                  "type": "change-packages-result"}])
        return result.addCallback(got_result)

    def test_perform_changes_with_allow_install_policy(self):
        """
        The C{POLICY_ALLOW_INSTALLS} policy the makes the changer mark
        the missing packages for installation.
        """
        self.store.set_hash_ids({HASH1: 1})
        self.facade.reload_channels()
        [package1] = self.facade.get_packages_by_name("name1")

        self.mocker.order()
        self.facade.perform_changes = self.mocker.mock()
        self.facade.perform_changes()
        self.mocker.throw(DependencyError([package1]))

        self.facade.mark_install = self.mocker.mock()
        self.facade.mark_install(package1)
        self.facade.perform_changes()
        self.mocker.result("success")
        self.mocker.replay()

        result = self.changer.change_packages(POLICY_ALLOW_INSTALLS)

        self.assertEqual(result.code, SUCCESS_RESULT)
        self.assertEqual(result.text, "success")
        self.assertEqual(result.installs, [1])
        self.assertEqual(result.removals, [])

    def test_perform_changes_with_allow_install_policy_and_removals(self):
        """
        The C{POLICY_ALLOW_INSTALLS} policy doesn't allow additional packages
        to be removed.
        """
        installed_hash = self.set_pkg1_installed()
        self.store.set_hash_ids({installed_hash: 1, HASH2: 2})
        self.facade.reload_channels()

        package1 = self.facade.get_package_by_hash(installed_hash)
        [package2] = self.facade.get_packages_by_name("name2")
        self.facade.perform_changes = self.mocker.mock()
        self.facade.perform_changes()
        self.mocker.throw(DependencyError([package1, package2]))
        self.mocker.replay()

        result = self.changer.change_packages(POLICY_ALLOW_INSTALLS)

        self.assertEqual(result.code, DEPENDENCY_ERROR_RESULT)
        self.assertEqual(result.text, None)
        self.assertEqual(result.installs, [2])
        self.assertEqual(result.removals, [1])

    def test_perform_changes_with_max_retries(self):
        """
        After having complemented the requested changes to handle a dependency
        error, the L{PackageChanger.change_packages} will try to perform the
        requested changes again only once.
        """
        self.store.set_hash_ids({HASH1: 1, HASH2: 2})
        self.facade.reload_channels()

        [package1] = self.facade.get_packages_by_name("name1")
        [package2] = self.facade.get_packages_by_name("name2")

        self.facade.perform_changes = self.mocker.mock()
        self.facade.perform_changes()
        self.mocker.throw(DependencyError([package1]))
        self.facade.perform_changes()
        self.mocker.throw(DependencyError([package2]))
        self.mocker.replay()

        result = self.changer.change_packages(POLICY_ALLOW_INSTALLS)

        self.assertEqual(result.code, DEPENDENCY_ERROR_RESULT)
        self.assertEqual(result.text, None)
        self.assertEqual(result.installs, [1, 2])
        self.assertEqual(result.removals, [])

    def test_handle_change_packages_with_policy(self):
        """
        The C{change-packages} message can have an optional C{policy}
        field that will be passed to the C{perform_changes} method.
        """
        self.store.set_hash_ids({HASH1: 1})
        self.store.add_task("changer",
                            {"type": "change-packages",
                             "install": [1],
                             "policy": POLICY_ALLOW_INSTALLS,
                             "operation-id": 123})
        self.changer.change_packages = self.mocker.mock()
        self.changer.change_packages(POLICY_ALLOW_INSTALLS)
        result = ChangePackagesResult()
        result.code = SUCCESS_RESULT
        self.mocker.result(result)
        self.mocker.replay()

        self.disable_clear_channels()
        return self.changer.handle_tasks()

    def test_perform_changes_with_policy_allow_all_changes(self):
        """
        The C{POLICY_ALLOW_ALL_CHANGES} policy allows any needed additional
        package to be installed or removed.
        """
        installed_hash = self.set_pkg1_installed()
        self.store.set_hash_ids({installed_hash: 1, HASH2: 2})
        self.facade.reload_channels()

        self.mocker.order()
        package1 = self.facade.get_package_by_hash(installed_hash)
        [package2] = self.facade.get_packages_by_name("name2")
        self.facade.perform_changes = self.mocker.mock()
        self.facade.perform_changes()
        self.mocker.throw(DependencyError([package1, package2]))
        self.facade.mark_install = self.mocker.mock()
        self.facade.mark_remove = self.mocker.mock()
        self.facade.mark_install(package2)
        self.facade.mark_remove(package1)
        self.facade.perform_changes()
        self.mocker.result("success")
        self.mocker.replay()

        result = self.changer.change_packages(POLICY_ALLOW_ALL_CHANGES)

        self.assertEqual(result.code, SUCCESS_RESULT)
        self.assertEqual(result.text, "success")
        self.assertEqual(result.installs, [2])
        self.assertEqual(result.removals, [1])

    def test_transaction_error(self):
        """
        In this case, the package we're trying to install declared some
        dependencies that can't be satisfied in the client because they
        don't exist at all.  The client must answer the request letting
        the server know about the unsolvable problem.
        """
        self.store.set_hash_ids({HASH1: 1})
        self.store.add_task("changer",
                            {"type": "change-packages", "install": [1],
                             "operation-id": 123})

        self.disable_clear_channels()
        result = self.changer.handle_tasks()

        def got_result(result):
            messages = self.get_pending_messages()
            self.assertEqual(len(messages), 1)
            message = messages[0]
            self.assertEqual(message["operation-id"], 123)
            self.assertEqual(message["result-code"], 100)
            self.assertIn(
                "packages have unmet dependencies", message["result-text"])
            self.assertEqual(message["type"], "change-packages-result")
        return result.addCallback(got_result)

    def test_tasks_are_isolated_marks(self):
        """
        Changes attempted on one task should be reset before the next
        task is run.  In this test, we try to run two different
        operations, first installing package 2, then upgrading
        anything available.  The first installation will fail for lack
        of superuser privileges, and the second one will succeed since
        there's nothing to upgrade.  If tasks are mixed up, the second
        operation will fail too, because the installation of package 2
        is still marked in the facade.
        """
        self.log_helper.ignore_errors(".*dpkg")

        installable_hash = self.set_pkg2_satisfied()
        installed_hash = self.set_pkg1_installed()
        self.store.set_hash_ids({installed_hash: 1, installable_hash: 2})

        self.store.add_task("changer",
                            {"type": "change-packages", "install": [2],
                             "operation-id": 123})
        self.store.add_task("changer",
                            {"type": "change-packages", "upgrade-all": True,
                             "operation-id": 124})

        result = self.changer.handle_tasks()

        def got_result(result):
            message = self.get_pending_messages()[1]
            self.assertEqual(124, message["operation-id"])
            self.assertEqual("change-packages-result", message["type"])
            self.assertNotEqual(0, message["result-code"])

        return result.addCallback(got_result)

    def test_tasks_are_isolated_cache(self):
        """
        The package (APT) cache should be reset between task runs.
        In this test, we try to run two different operations, first
        installing package 2, then removing package 1.  Both tasks will
        fail for lack of superuser privileges.  If the package cache
        isn't reset between tasks, the second operation will fail with a
        dependency error, since it will be marked for installation, but
        we haven't explicitly marked it so.
        """
        self.log_helper.ignore_errors(".*dpkg")

        installable_hash = self.set_pkg2_satisfied()
        installed_hash = self.set_pkg1_installed()
        self.store.set_hash_ids({installed_hash: 1, installable_hash: 2})

        self.store.add_task("changer",
                            {"type": "change-packages", "install": [2],
                             "operation-id": 123})
        self.store.add_task("changer",
                            {"type": "change-packages", "remove": [1],
                             "operation-id": 124})

        result = self.changer.handle_tasks()

        def got_result(result):
            message1, message2 = self.get_pending_messages()
            self.assertEqual(123, message1["operation-id"])
            self.assertEqual("change-packages-result", message1["type"])
            self.assertEqual(ERROR_RESULT, message1["result-code"])
            self.assertEqual(124, message2["operation-id"])
            self.assertEqual("change-packages-result", message2["type"])
            self.assertEqual(ERROR_RESULT, message2["result-code"])

        return result.addCallback(got_result)

    def test_successful_operation(self):
        """Simulate a *very* successful operation.

        We'll do that by hacking perform_changes(), and returning our
        *very* successful operation result.
        """
        installed_hash = self.set_pkg1_installed()
        self.store.set_hash_ids({installed_hash: 1, HASH2: 2, HASH3: 3})
        self.store.add_task("changer",
                            {"type": "change-packages", "install": [2],
                             "operation-id": 123})

        def return_good_result(self):
            return "Yeah, I did whatever you've asked for!"
        self.replace_perform_changes(return_good_result)

        result = self.changer.handle_tasks()

        def got_result(result):
            self.assertMessages(self.get_pending_messages(),
                                [{"operation-id": 123,
                                  "result-code": 1,
                                  "result-text": "Yeah, I did whatever you've "
                                                 "asked for!",
                                  "type": "change-packages-result"}])
        return result.addCallback(got_result)

    def test_successful_operation_with_binaries(self):
        """
        Simulate a successful operation involving server-generated binary
        packages.
        """
        self.store.set_hash_ids({HASH3: 3})
        self.store.add_task("changer",
                            {"type": "change-packages", "install": [2, 3],
                             "binaries": [(HASH2, 2, PKGDEB2)],
                             "operation-id": 123})

        def return_good_result(self):
            return "Yeah, I did whatever you've asked for!"
        self.replace_perform_changes(return_good_result)
        self.disable_clear_channels()

        result = self.changer.handle_tasks()

        def got_result(result):
            self.assertMessages(self.get_pending_messages(),
                                [{"operation-id": 123,
                                  "result-code": 1,
                                  "result-text": "Yeah, I did whatever you've "
                                                 "asked for!",
                                  "type": "change-packages-result"}])
        return result.addCallback(got_result)

    def test_global_upgrade(self):
        """
        Besides asking for individual changes, the server may also request
        the client to perform a global upgrade.  This would be the equivalent
        of a "apt-get upgrade" command being executed in the command line.
        """
        hash1, hash2 = self.set_pkg2_upgrades_pkg1()
        self.store.set_hash_ids({hash1: 1, hash2: 2})

        self.store.add_task("changer",
                            {"type": "change-packages", "upgrade-all": True,
                             "operation-id": 123})

        result = self.changer.handle_tasks()

        def got_result(result):
            self.assertMessages(self.get_pending_messages(),
                                [{"operation-id": 123,
                                  "must-install": [2],
                                  "must-remove": [1],
                                  "result-code": 101,
                                  "type": "change-packages-result"}])

        return result.addCallback(got_result)

    def test_global_upgrade_with_nothing_to_do(self):

        self.store.add_task("changer",
                            {"type": "change-packages", "upgrade-all": True,
                             "operation-id": 123})

        result = self.changer.handle_tasks()

        def got_result(result):
            self.assertMessages(self.get_pending_messages(),
                                [{"operation-id": 123,
                                  "result-code": 1,
                                  "type": "change-packages-result"}])

        return result.addCallback(got_result)

    def test_run_with_no_update_stamp(self):
        """
        If the update-stamp file is not there yet, the package changer
        just exists.
        """
        os.remove(self.config.update_stamp_filename)

        def assert_log(ignored):
            self.assertIn("The package-reporter hasn't run yet, exiting.",
                          self.logfile.getvalue())

        result = self.changer.run()
        return result.addCallback(assert_log)

    def test_spawn_reporter_after_running(self):
        output_filename = self.makeFile("REPORTER NOT RUN")
        reporter_filename = self.makeFile("#!/bin/sh\necho REPORTER RUN > %s" %
                                          output_filename)
        os.chmod(reporter_filename, 0755)

        find_command_mock = self.mocker.replace(
            "landscape.package.reporter.find_reporter_command")
        find_command_mock()
        self.mocker.result(reporter_filename)
        self.mocker.replay()

        # Add a task that will do nothing besides producing an answer.
        # The reporter is only spawned if at least one task was handled.
        self.store.add_task("changer", {"type": "change-packages",
                                        "operation-id": 123})

        result = self.changer.run()

        def got_result(result):
            self.assertEqual(open(output_filename).read().strip(),
                             "REPORTER RUN")
        return result.addCallback(got_result)

    def test_spawn_reporter_after_running_with_config(self):
        """The changer passes the config to the reporter when running it."""
        self.config.config = "test.conf"
        output_filename = self.makeFile("REPORTER NOT RUN")
        reporter_filename = self.makeFile("#!/bin/sh\necho ARGS $@ > %s" %
                                          output_filename)
        os.chmod(reporter_filename, 0755)

        find_command_mock = self.mocker.replace(
            "landscape.package.reporter.find_reporter_command")
        find_command_mock()
        self.mocker.result(reporter_filename)
        self.mocker.replay()

        # Add a task that will do nothing besides producing an answer.
        # The reporter is only spawned if at least one task was handled.
        self.store.add_task("changer", {"type": "change-packages",
                                        "operation-id": 123})

        result = self.changer.run()

        def got_result(result):
            self.assertEqual(open(output_filename).read().strip(),
                             "ARGS -c test.conf")
        return result.addCallback(got_result)

    def test_set_effective_uid_and_gid_when_running_as_root(self):
        """
        After the package changer has run, we want the package-reporter to run
        to report the recent changes.  If we're running as root, we want to
        change to the "landscape" user and "landscape" group.
        """

        # We are running as root
        getuid_mock = self.mocker.replace("os.getuid")
        getuid_mock()
        self.mocker.result(0)

        self.mocker.order()

        # We want to return a known gid
        grnam_mock = self.mocker.replace("grp.getgrnam")
        grnam_mock("landscape")

        class FakeGroup(object):
            gr_gid = 199

        self.mocker.result(FakeGroup())

        # First the changer should change the group
        setgid_mock = self.mocker.replace("os.setgid")
        setgid_mock(199)

        # And a known uid as well
        pwnam_mock = self.mocker.replace("pwd.getpwnam")
        pwnam_mock("landscape")

        class FakeUser(object):
            pw_uid = 199

        self.mocker.result(FakeUser())

        # And now the user as well
        setuid_mock = self.mocker.replace("os.setuid")
        setuid_mock(199)

        # Finally, we don't really want the package reporter to run.
        system_mock = self.mocker.replace("os.system")
        system_mock(ANY)

        self.mocker.replay()

        # Add a task that will do nothing besides producing an answer.
        # The reporter is only spawned if at least one task was handled.
        self.store.add_task("changer", {"type": "change-packages",
                                        "operation-id": 123})
        return self.changer.run()

    def test_run(self):
        changer_mock = self.mocker.patch(self.changer)

        self.mocker.order()

        results = [Deferred() for i in range(2)]

        changer_mock.use_hash_id_db()
        self.mocker.result(results[0])

        changer_mock.handle_tasks()
        self.mocker.result(results[1])

        self.mocker.replay()

        self.changer.run()

        # It must raise an error because deferreds weren't yet fired.
        self.assertRaises(AssertionError, self.mocker.verify)

        for deferred in reversed(results):
            deferred.callback(None)

    def test_dont_spawn_reporter_after_running_if_nothing_done(self):
        output_filename = self.makeFile("REPORTER NOT RUN")
        reporter_filename = self.makeFile("#!/bin/sh\necho REPORTER RUN > %s" %
                                          output_filename)
        os.chmod(reporter_filename, 0755)

        find_command_mock = self.mocker.replace(
            "landscape.package.reporter.find_reporter_command")
        find_command_mock()
        self.mocker.result(reporter_filename)
        self.mocker.count(0, None)
        self.mocker.replay()

        result = self.changer.run()

        def got_result(result):
            self.assertEqual(open(output_filename).read().strip(),
                             "REPORTER NOT RUN")
        return result.addCallback(got_result)

    def test_main(self):
        self.mocker.order()

        run_task_handler = self.mocker.replace("landscape.package.taskhandler"
                                               ".run_task_handler",
                                               passthrough=False)
        getpgrp = self.mocker.replace("os.getpgrp")
        self.expect(getpgrp()).result(os.getpid() + 1)
        setsid = self.mocker.replace("os.setsid")
        setsid()
        run_task_handler(PackageChanger, ["ARGS"])
        self.mocker.result("RESULT")

        self.mocker.replay()

        self.assertEqual(main(["ARGS"]), "RESULT")

    def test_main_run_from_shell(self):
        """
        We want the getpid and getpgrp to return the same process id
        this simulates the case where the process is already the process
        session leader, in this case the os.setsid would fail.
        """
        getpgrp = self.mocker.replace("os.getpgrp")
        getpgrp()
        self.mocker.result(os.getpid())

        setsid = self.mocker.replace("os.setsid")
        setsid()
        self.mocker.count(0, 0)

        run_task_handler = self.mocker.replace("landscape.package.taskhandler"
                                               ".run_task_handler",
                                               passthrough=False)
        run_task_handler(PackageChanger, ["ARGS"])
        self.mocker.replay()

        main(["ARGS"])

    def test_find_changer_command(self):
        dirname = self.makeDir()
        filename = self.makeFile("", dirname=dirname,
                                 basename="landscape-package-changer")

        saved_argv = sys.argv
        try:
            sys.argv = [os.path.join(dirname, "landscape-monitor")]

            command = find_changer_command()

            self.assertEqual(command, filename)
        finally:
            sys.argv = saved_argv

    def test_transaction_error_with_unicode_data(self):
        self.store.set_hash_ids({HASH1: 1})
        self.store.add_task("changer",
                            {"type": "change-packages", "install": [1],
                             "operation-id": 123})

        def raise_error(self):
            raise TransactionError(u"áéíóú")
        self.replace_perform_changes(raise_error)
        self.disable_clear_channels()

        result = self.changer.handle_tasks()

        def got_result(result):
            self.assertMessages(self.get_pending_messages(),
                                [{"operation-id": 123,
                                  "result-code": 100,
                                  "result-text": u"áéíóú",
                                  "type": "change-packages-result"}])
        return result.addCallback(got_result)

    def test_update_stamp_exists(self):
        """
        L{PackageChanger.update_stamp_exists} returns C{True} if the
        update-stamp file is there, C{False} otherwise.
        """
        self.assertTrue(self.changer.update_stamp_exists())
        os.remove(self.config.update_stamp_filename)
        self.assertFalse(self.changer.update_stamp_exists())

    def test_update_stamp_exists_notifier(self):
        """
        L{PackageChanger.update_stamp_exists} also checks the existence of the
        C{update_notifier_stamp} file.
        """
        self.assertTrue(self.changer.update_stamp_exists())
        os.remove(self.config.update_stamp_filename)
        self.assertFalse(self.changer.update_stamp_exists())
        self.changer.update_notifier_stamp = self.makeFile("")
        self.assertTrue(self.changer.update_stamp_exists())

    def test_binaries_path(self):
        self.assertEqual(
            self.config.binaries_path,
            os.path.join(self.config.data_path, "package", "binaries"))

    def test_init_channels(self):
        """
        The L{PackageChanger.init_channels} method makes the given
        Debian packages available in a facade channel.
        """
        binaries = [(HASH1, 111, PKGDEB1), (HASH2, 222, PKGDEB2)]

        self.facade.reset_channels()
        self.changer.init_channels(binaries)

        binaries_path = self.config.binaries_path
        self.assertFileContent(os.path.join(binaries_path, "111.deb"),
                               base64.decodestring(PKGDEB1))
        self.assertFileContent(os.path.join(binaries_path, "222.deb"),
                               base64.decodestring(PKGDEB2))
        self.assertEqual(
            self.facade.get_channels(),
            self.get_binaries_channels(binaries_path))

        self.assertEqual(self.store.get_hash_ids(), {HASH1: 111, HASH2: 222})

        self.facade.ensure_channels_reloaded()
        [pkg1, pkg2] = sorted(self.facade.get_packages(),
                              key=self.get_package_name)
        self.assertEqual(self.facade.get_package_hash(pkg1), HASH1)
        self.assertEqual(self.facade.get_package_hash(pkg2), HASH2)

    def test_init_channels_with_existing_hash_id_map(self):
        """
        The L{PackageChanger.init_channels} behaves well even if the
        hash->id mapping for a given deb is already in the L{PackageStore}.
        """
        self.store.set_hash_ids({HASH1: 111})
        self.changer.init_channels([(HASH1, 111, PKGDEB1)])
        self.assertEqual(self.store.get_hash_ids(), {HASH1: 111})

    def test_init_channels_with_existing_binaries(self):
        """
        The L{PackageChanger.init_channels} removes Debian packages
        from previous runs.
        """
        existing_deb_path = os.path.join(self.config.binaries_path, "123.deb")
        self.makeFile(basename=existing_deb_path, content="foo")
        self.changer.init_channels([])
        self.assertFalse(os.path.exists(existing_deb_path))

    def test_binaries_available_in_cache(self):
        """
        If binaries are included in the changes-packages message, those
        will be added to the facade's cache.
        """
        # Make sure to turn off automatic rereading of Packages file,
        # like it is by default.
        self.facade.refetch_package_index = False
        self.assertEqual(None, self.facade.get_package_by_hash(HASH2))
        self.store.add_task("changer",
                            {"type": "change-packages", "install": [2],
                             "binaries": [(HASH2, 2, PKGDEB2)],
                             "operation-id": 123})

        def return_good_result(self):
            return "Yeah, I did whatever you've asked for!"
        self.replace_perform_changes(return_good_result)

        result = self.changer.handle_tasks()

        def got_result(result):
            self.assertNotEqual(None, self.facade.get_package_by_hash(HASH2))
            self.assertFalse(self.facade.refetch_package_index)

        return result.addCallback(got_result)

    def test_change_package_holds(self):
        """
        The L{PackageChanger.handle_tasks} method appropriately creates and
        deletes package holds as requested by the C{change-packages}
        message.
        """
        self._add_system_package("foo")
        self._add_system_package("bar")
        self.facade.reload_channels()
        self._hash_packages_by_name(self.facade, self.store, "foo")
        self._hash_packages_by_name(self.facade, self.store, "bar")
        [foo] = self.facade.get_packages_by_name("foo")
        [bar] = self.facade.get_packages_by_name("bar")
        self.facade.set_package_hold(bar)
        # Make sure that the mtime of the dpkg status file is old when
        # apt loads it, so that it will be reloaded when asserting the
        # test result.
        old_mtime = time.time() - 10
        os.utime(self.facade._dpkg_status, (old_mtime, old_mtime))
        self.facade.reload_channels()
        self.store.add_task("changer", {"type": "change-packages",
                                        "hold": [foo.package.id],
                                        "remove-hold": [bar.package.id],
                                        "operation-id": 123})

        def assert_result(result):
            self.facade.reload_channels()
            self.assertEqual(["foo"], self.facade.get_package_holds())
            self.assertIn("Queuing response with change package results "
                          "to exchange urgently.", self.logfile.getvalue())
            self.assertMessages(
                self.get_pending_messages(),
                [{"type": "change-packages-result",
                  "operation-id": 123,
                  "result-text": "Package holds successfully changed.",
                  "result-code": 1}])

        result = self.changer.handle_tasks()
        return result.addCallback(assert_result)

    def test_create_package_holds_with_identical_version(self):
        """
        The L{PackageChanger.handle_tasks} method appropriately creates
        holds as requested by the C{change-packages} message even
        when versions from two different packages are the same.
        """
        self._add_system_package("foo", version="1.1")
        self._add_system_package("bar", version="1.1")
        self.facade.reload_channels()
        self._hash_packages_by_name(self.facade, self.store, "foo")
        self._hash_packages_by_name(self.facade, self.store, "bar")
        [foo] = self.facade.get_packages_by_name("foo")
        [bar] = self.facade.get_packages_by_name("bar")
        self.facade.reload_channels()
        self.store.add_task("changer", {"type": "change-packages",
                                        "hold": [foo.package.id,
                                                         bar.package.id],
                                        "operation-id": 123})

        def assert_result(result):
            self.facade.reload_channels()
            self.assertEqual(["foo", "bar"], self.facade.get_package_holds())

        result = self.changer.handle_tasks()
        return result.addCallback(assert_result)

    def test_delete_package_holds_with_identical_version(self):
        """
        The L{PackageChanger.handle_tasks} method appropriately deletes
        holds as requested by the C{change-packages} message even
        when versions from two different packages are the same.
        """
        self._add_system_package("foo", version="1.1")
        self._add_system_package("bar", version="1.1")
        self.facade.reload_channels()
        self._hash_packages_by_name(self.facade, self.store, "foo")
        self._hash_packages_by_name(self.facade, self.store, "bar")
        [foo] = self.facade.get_packages_by_name("foo")
        [bar] = self.facade.get_packages_by_name("bar")
        self.facade.set_package_hold(foo)
        self.facade.set_package_hold(bar)
        self.facade.reload_channels()
        self.store.add_task("changer", {"type": "change-packages",
                                        "remove-hold": [foo.package.id,
                                                         bar.package.id],
                                        "operation-id": 123})

        def assert_result(result):
            self.facade.reload_channels()
            self.assertEqual([], self.facade.get_package_holds())

        result = self.changer.handle_tasks()
        return result.addCallback(assert_result)

    def test_change_package_holds_create_already_held(self):
        """
        If the C{change-packages} message requests to add holds for
        packages that are already held, the activity succeeds, since the
        end result is that the requested package holds are there.
        """
        self._add_system_package("foo")
        self.facade.reload_channels()
        self._hash_packages_by_name(self.facade, self.store, "foo")
        [foo] = self.facade.get_packages_by_name("foo")
        self.facade.set_package_hold(foo)
        self.facade.reload_channels()
        self.store.add_task("changer", {"type": "change-packages",
                                        "hold": [foo.package.id],
                                        "operation-id": 123})

        def assert_result(result):
            self.facade.reload_channels()
            self.assertEqual(["foo"], self.facade.get_package_holds())
            self.assertIn("Queuing response with change package results "
                          "to exchange urgently.", self.logfile.getvalue())
            self.assertMessages(
                self.get_pending_messages(),
                [{"type": "change-packages-result",
                  "operation-id": 123,
                  "result-text": "Package holds successfully changed.",
                  "result-code": 1}])

        result = self.changer.handle_tasks()
        return result.addCallback(assert_result)

    def test_change_package_holds_create_other_version_installed(self):
        """
        If the C{change-packages} message requests to add holds for
        packages that have a different version installed than the one
        being requested to hold, the activity fails.

        The whole activity is failed, meaning that other valid hold
        requests won't get processed.
        """
        self._add_system_package("foo", version="1.0")
        self._add_package_to_deb_dir(
            self.repository_dir, "foo", version="2.0")
        self._add_system_package("bar", version="1.0")
        self._add_package_to_deb_dir(
            self.repository_dir, "bar", version="2.0")
        self.facade.reload_channels()
        [foo1, foo2] = sorted(self.facade.get_packages_by_name("foo"))
        [bar1, bar2] = sorted(self.facade.get_packages_by_name("bar"))
        self.store.set_hash_ids({self.facade.get_package_hash(foo1): 1,
                                 self.facade.get_package_hash(foo2): 2,
                                 self.facade.get_package_hash(bar1): 3,
                                 self.facade.get_package_hash(bar2): 4})
        self.facade.reload_channels()
        self.store.add_task("changer", {"type": "change-packages",
                                        "hold": [2, 3],
                                        "operation-id": 123})

        def assert_result(result):
            self.facade.reload_channels()
            self.assertEqual([], self.facade.get_package_holds())
            self.assertIn("Queuing response with change package results "
                          "to exchange urgently.", self.logfile.getvalue())
            self.assertMessages(
                self.get_pending_messages(),
                [{"type": "change-packages-result",
                  "operation-id": 123,
                  "result-text": "Cannot perform the changes, since the" +
                                 " following packages are not installed: foo",
                  "result-code": 100}])

        result = self.changer.handle_tasks()
        return result.addCallback(assert_result)

    def test_change_package_holds_create_not_installed(self):
        """
        If the C{change-packages} message requests to add holds for
        packages that aren't installed, the whole activity is failed. If
        multiple holds are specified, those won't be added. There's no
        difference between a package that is available in some
        repository and a package that the facade doesn't know about at
        all.
        """
        self._add_system_package("foo")
        self._add_package_to_deb_dir(self.repository_dir, "bar")
        self._add_package_to_deb_dir(self.repository_dir, "baz")
        self.facade.reload_channels()
        self._hash_packages_by_name(self.facade, self.store, "foo")
        self._hash_packages_by_name(self.facade, self.store, "bar")
        self._hash_packages_by_name(self.facade, self.store, "baz")
        [foo] = self.facade.get_packages_by_name("foo")
        [bar] = self.facade.get_packages_by_name("bar")
        [baz] = self.facade.get_packages_by_name("baz")
        self.store.add_task("changer", {"type": "change-packages",
                                        "hold": [foo.package.id,
                                                         bar.package.id,
                                                         baz.package.id],
                                        "operation-id": 123})

        def assert_result(result):
            self.facade.reload_channels()
            self.assertEqual([], self.facade.get_package_holds())
            self.assertIn("Queuing response with change package results "
                          "to exchange urgently.", self.logfile.getvalue())
            self.assertMessages(
                self.get_pending_messages(),
                [{"type": "change-packages-result",
                  "operation-id": 123,
                  "result-text": "Cannot perform the changes, since the "
                  "following packages are not installed: "
                  "%s, %s" % tuple(sorted([bar.package.name,
                                           baz.package.name])),
                  "result-code": 100}])

        result = self.changer.handle_tasks()
        return result.addCallback(assert_result)

    def test_change_package_holds_create_unknown_hash(self):
        """
        If the C{change-packages} message requests to add holds for
        packages that the client doesn't know about results in a not yet
        synchronized message and a failure of the operation.
        """

        self.store.add_task("changer",
                            {"type": "change-packages",
                             "hold": [123],
                             "operation-id": 123})

        time_mock = self.mocker.replace("time.time")
        time_mock()
        self.mocker.result(time.time() + UNKNOWN_PACKAGE_DATA_TIMEOUT)
        self.mocker.count(1, None)
        self.mocker.replay()

        try:
            result = self.changer.handle_tasks()
            self.mocker.verify()
        finally:
            # Reset it earlier so that Twisted has the true time function.
            self.mocker.reset()

        self.assertIn("Package data not yet synchronized with server (123)",
                      self.logfile.getvalue())

        def got_result(result):
            message = {"type": "change-packages-result",
                       "operation-id": 123,
                       "result-code": 100,
                       "result-text": "Package data has changed. "
                                      "Please retry the operation."}
            self.assertMessages(self.get_pending_messages(), [message])
            self.assertEqual(self.store.get_next_task("changer"), None)
        return result.addCallback(got_result)

    def test_change_package_holds_delete_not_held(self):
        """
        If the C{change-packages} message requests to remove holds
        for packages that aren't held, the activity succeeds if the
        right version is installed, since the end result is that the
        hold is removed.
        """
        self._add_package_to_deb_dir(self.repository_dir, "foo")
        self.facade.reload_channels()
        self._hash_packages_by_name(self.facade, self.store, "foo")
        [foo] = self.facade.get_packages_by_name("foo")
        self.store.add_task("changer", {"type": "change-packages",
                                        "remove-hold": [foo.package.id],
                                        "operation-id": 123})

        def assert_result(result):
            self.facade.reload_channels()
            self.assertEqual([], self.facade.get_package_holds())
            self.assertIn("Queuing response with change package results "
                          "to exchange urgently.", self.logfile.getvalue())
            self.assertMessages(
                self.get_pending_messages(),
                [{"type": "change-packages-result",
                  "operation-id": 123,
                  "result-text": "Package holds successfully changed.",
                  "result-code": 1}])

        result = self.changer.handle_tasks()
        return result.addCallback(assert_result)

    def test_change_package_holds_delete_different_version_held(self):
        """
        If the C{change-packages} message requests to remove holds
        for packages that aren't held, the activity succeeds if the
        right version is installed, since the end result is that the
        hold is removed.
        """
        self._add_system_package("foo", version="1.0")
        self._add_package_to_deb_dir(
            self.repository_dir, "foo", version="2.0")
        self.facade.reload_channels()
        [foo1, foo2] = sorted(self.facade.get_packages_by_name("foo"))
        self.store.set_hash_ids({self.facade.get_package_hash(foo1): 1,
                                 self.facade.get_package_hash(foo2): 2})
        self.facade.mark_install(foo1)
        self.facade.mark_hold(foo1)
        self.facade.perform_changes()
        self.facade.reload_channels()
        self.store.add_task("changer", {"type": "change-packages",
                                        "remove-hold": [2],
                                        "operation-id": 123})

        def assert_result(result):
            self.facade.reload_channels()
            self.assertEqual(["foo"], self.facade.get_package_holds())
            self.assertIn("Queuing response with change package results "
                          "to exchange urgently.", self.logfile.getvalue())
            self.assertMessages(
                self.get_pending_messages(),
                [{"type": "change-packages-result",
                  "operation-id": 123,
                  "result-text": "Package holds successfully changed.",
                  "result-code": 1}])

        result = self.changer.handle_tasks()
        return result.addCallback(assert_result)

    def test_change_package_holds_delete_not_installed(self):
        """
        If the C{change-packages} message requests to remove holds
        for packages that aren't installed, the activity succeeds, since
        the end result is still that the package isn't held at the
        requested version.
        """
        self._add_system_package("foo")
        self.facade.reload_channels()
        self._hash_packages_by_name(self.facade, self.store, "foo")
        [foo] = self.facade.get_packages_by_name("foo")
        self.store.add_task("changer", {"type": "change-packages",
                                        "remove-hold": [foo.package.id],
                                        "operation-id": 123})

        def assert_result(result):
            self.facade.reload_channels()
            self.assertEqual([], self.facade.get_package_holds())
            self.assertIn("Queuing response with change package results "
                          "to exchange urgently.", self.logfile.getvalue())
            self.assertMessages(
                self.get_pending_messages(),
                [{"type": "change-packages-result",
                  "operation-id": 123,
                  "result-text": "Package holds successfully changed.",
                  "result-code": 1}])

        result = self.changer.handle_tasks()
        return result.addCallback(assert_result)

    def test_change_package_locks(self):
        """
        The L{PackageChanger.handle_tasks} method fails
        change-package-locks activities, since it can't add or remove
        locks because apt doesn't support this.
        """
        self.store.add_task("changer", {"type": "change-package-locks",
                                        "create": [("foo", ">=", "1.0")],
                                        "delete": [("bar", None, None)],
                                        "operation-id": 123})

        def assert_result(result):
            self.assertMessages(
                self.get_pending_messages(),
                [{"type": "operation-result",
                  "operation-id": 123,
                  "status": FAILED,
                  "result-text": "This client doesn't support package locks.",
                  "result-code": 1}])

        result = self.changer.handle_tasks()
        return result.addCallback(assert_result)

    def test_change_packages_with_binaries_removes_binaries(self):
        """
        After the C{change-packages} handler has installed the binaries,
        the binaries and the internal facade deb source is removed.
        """
        self.store.add_task("changer",
                            {"type": "change-packages", "install": [2],
                             "binaries": [(HASH2, 2, PKGDEB2)],
                             "operation-id": 123})

        def return_good_result(self):
            return "Yeah, I did whatever you've asked for!"
        self.replace_perform_changes(return_good_result)

        result = self.changer.handle_tasks()

        def got_result(result):
            self.assertMessages(self.get_pending_messages(),
                                [{"operation-id": 123,
                                  "result-code": 1,
                                  "result-text": "Yeah, I did whatever you've "
                                                 "asked for!",
                                  "type": "change-packages-result"}])
            self.assertEqual([], os.listdir(self.config.binaries_path))
            self.assertFalse(
                os.path.exists(self.facade._get_internal_sources_list()))

        return result.addCallback(got_result)

    def test_change_packages_with_reboot_flag(self):
        """
        When a C{reboot-if-necessary} flag is passed in the C{change-packages},
        A C{ShutdownProtocolProcess} is created and the package result change
        is returned.
        """
        self.store.add_task("changer",
                            {"type": "change-packages", "install": [2],
                             "binaries": [(HASH2, 2, PKGDEB2)],
                             "operation-id": 123,
                             "reboot-if-necessary": True})

        def return_good_result(self):
            return "Yeah, I did whatever you've asked for!"
        self.replace_perform_changes(return_good_result)

        result = self.changer.handle_tasks()

        def got_result(result):
            self.assertIn("Landscape is rebooting the system",
                          self.logfile.getvalue())
            self.assertMessages(self.get_pending_messages(),
                                [{"operation-id": 123,
                                  "result-code": 1,
                                  "result-text": "Yeah, I did whatever you've "
                                                 "asked for!",
                                  "type": "change-packages-result"}])

        self.landscape_reactor.advance(5)
        [arguments] = self.process_factory.spawns
        protocol = arguments[0]
        protocol.processEnded(Failure(ProcessDone(status=0)))
        self.broker_service.reactor.advance(100)
        self.landscape_reactor.advance(10)
        return result.addCallback(got_result)

    def test_change_packages_with_failed_reboot(self):
        """
        When a C{reboot-if-necessary} flag is passed in the C{change-packages},
        A C{ShutdownProtocol} is created and the package result change is
        returned, even if the reboot fails.
        """
        self.store.add_task("changer",
                            {"type": "change-packages", "install": [2],
                             "binaries": [(HASH2, 2, PKGDEB2)],
                             "operation-id": 123,
                             "reboot-if-necessary": True})

        def return_good_result(self):
            return "Yeah, I did whatever you've asked for!"
        self.replace_perform_changes(return_good_result)

        result = self.changer.handle_tasks()

        def got_result(result):
            self.assertMessages(self.get_pending_messages(),
                                [{"operation-id": 123,
                                  "result-code": 1,
                                  "result-text": "Yeah, I did whatever you've "
                                                 "asked for!",
                                  "type": "change-packages-result"}])
            self.log_helper.ignore_errors(ShutdownFailedError)

        self.landscape_reactor.advance(5)
        [arguments] = self.process_factory.spawns
        protocol = arguments[0]
        protocol.processEnded(Failure(ProcessTerminated(exitCode=1)))
        self.landscape_reactor.advance(10)
        return result.addCallback(got_result)

    def test_no_exchange_after_reboot(self):
        """
        After initiating a reboot process, no more messages are exchanged.
        """
        self.store.add_task("changer",
                            {"type": "change-packages", "install": [2],
                             "binaries": [(HASH2, 2, PKGDEB2)],
                             "operation-id": 123,
                             "reboot-if-necessary": True})

        def return_good_result(self):
            return "Yeah, I did whatever you've asked for!"
        self.replace_perform_changes(return_good_result)

        result = self.changer.handle_tasks()

        def got_result(result):
            # Advance both reactors so the pending messages are exchanged.
            self.broker_service.reactor.advance(100)
            self.landscape_reactor.advance(10)
            payloads = self.broker_service.exchanger._transport.payloads
            self.assertEqual(0, len(payloads))

        self.landscape_reactor.advance(5)

        [arguments] = self.process_factory.spawns
        protocol = arguments[0]
        protocol.processEnded(Failure(ProcessDone(status=0)))
        self.broker_service.reactor.advance(100)
        self.landscape_reactor.advance(10)
        return result.addCallback(got_result)

    def test_run_gets_session_id(self):
        """
        Invoking L{PackageChanger.run} results in the session ID being fetched.
        """
        def assert_session_id(ignored):
            self.assertTrue(self.changer._session_id is not None)

        self.changer._session_id = None
        result = self.changer.run()
        return result.addCallback(assert_session_id)
